/*
 *   WDX: WitchDoctor-WDX, versatile distributed lock manager
 *   Copyright (C) 2003 Dominique Chabord,
 *   Please refer to credits section below
 *
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   ************************************************
 *
 *   History and credits:
 *   First version of WitchDoctor-WDX was based on:
 *
 *   DECwatchdog/Autopilot, watchdog and lock manager
 *   Copyright (C) 2003 HP
 *
 *   Developed by
 *     Computer Special Systems, Annecy France
 *     A Division of Digital Equipment Corporation
 *   and
 *     Custom-Systems, Annecy France
 *     An engineering center of Compaq
 *
 *   Provided under GNU-GPL terms by HP, 30 June 2003
 *   GNU-GPL replaces previous license statements
 *
 *   Rebranded, ported and modified for Shaman-X project, 15 July 2003
 *   by Dominique Chabord
 *   and maintained by other contributors as mentionned in the code
 *
 *   ************************************************
 *
 *   WDX is promoted by Shaman-X project
 *
 *   Check for new versions and documentation at
 *   http://www.shaman-x.org
 *
 *   ************************************************
 *
 *   Most of WDX code is designed to run unchanged on
 *   different operating systems, including at V1.0 time:
 *   Gnu-Linux, WindowsNT, DigitalUNIX and OpenVMS.
 *   Operating system specifics are identified by dedicated files.
 *
 *   Evolutions
 *   ----------
 *   July-2003     Dominique Chabord    WitchDoctor-WDX V1.0
 *                 Creation
 *
 *   ************************************************
 *
 *
 *	wdx_kernel.c
 *
 *	Provide wdx process services
 *
 *      Routines :
 *
 *	Public Routines :
 *      void kernel_init()
 *      int kernel_scheduler()
 *
 *	Local Routines
 *
 *	void activity_protocol()
 *	void reset_command_wish()
 *      unsigned int build_data_number()
 *      int validate_data_number()
 *      void reset_application_slot()
 *	int find_appplication_slot()
 *      int send_acknowledge()
 *	void create_asynchronous_message()
 *      int is_checksum_valid()
 *      int is_message_valid()
 *
 *	Command message management routines :
 *      int validate_silence()
 *      enum command_answer answer_to_start_request()
 *      void delete_delayed_operation()
 *	int exclude_protocol()
 *	int exclude_act()
 *	int conflict_act()
 *	int ignore_act()
 *	int answer_act()
 *	int process_act()
 *
 *	Operation message management routines :
 *	int action_opr()
 *	int reject_opr()
 *	int request_opr()
 *	int local_opr()
 *	int wait_opr()
 *	int master_opr()
 *	int masdel_opr()
 *	int delay_opr()
 *	int lock_opr()
 *      int control_opr()
 *
 *	Switch for incoming messages
 *      int command_message()
 *      int process_message()
 *
 ************************************************************************/
/*
 *
 *  11-Feb-94 -- Cornu
 *      Creation
 *
 * MK001 - added parameter wdxsend to wat_local_member_transition
 */

/* DM 14/09/03 replace 'ftime' with 'gettimeofday()' */
/* DM 21/08/03 Added previous changes CDM001, CDM002, CDM003 */
/*
 * CDM001 - 14-MAY-96  Fix a bug . Now, the function "ftime" is called
 *                     only one time in ker_scheduler loop
 *
 * CDM002 - 15-JUL-96  Add a test to detect if the broadcast time has been
 *                     exceeded between 2 sending. If that is the case, a
 *                     message is logged.
 *                     The test is implemented only if ACTIVITY (support
 *                     parameter) is ENABLED
 *
 * CDM003 - 05-AUG-96  Add a local correction for the time used to check
 *                     Operating System scheduling (this correction is
 *                     necessary for display in case of problem).
 */

/* Abstract :
 *
 * To initialize this module the 'ker_init' routine must be called first
 *
 */
#undef  WDX_KERNEL_MODULE
#define WDX_KERNEL_MODULE
#include	"wdx.h"
#include	"wdx_kernel.h"

static enum  watchdog_status       status; /* status of the kernel process */
static enum  watchdog_substatus substatus; /* substatus of the kernel process */
static struct command_wish        cmdwish; /* command wish data structure */
static int		  	  silence; /* silence delay value */
static int		  	  exclude; /* exclude delay value */
static int			  wdxsend; /* flag used to define if answer is to be sent */
/* if 0 no resource written in hearbeat / if 1 heartbeat updated */
static char			     full;
/* activity problem detected by activity algorithm */
static char		  activityproblem;


/*
 * PUBLIC ROUTINES : prefixe is "ker"
 */

/* void ker_init()
 * Init static variables used by kernel module
 *
 * Argument : COLD_INIT or WARM_INIT mode
 *	      pt to group database
 *            pt to message structure
 *
 */
void ker_init(s,gp,mp)
enum startup_mode   s; /* startup mode */
struct group      *gp; /* group pointer */
struct message    *mp; /* new message pointer */
{
int i;  /* index */
int j;  /* index */

   /* generic init */
   gdp = gp;
   localmemberp = &(gp->memberlist[localindex]);
   mess = mp;

   /* silence init : max of broadcast time + 1 */
	 silence = 0;
	 for(i=0;i<gdp->membercount;i++)
		 if(gdp->memberlist[i].broadcast > silence )
			  silence = gdp->memberlist[i].broadcast;
	 silence++;

   /* exclude init : max of broadcast time + 1 */
   exclude = 0;
   for(i=0;i<gdp->membercount;i++)
	   if(gdp->memberlist[i].broadcast > exclude )
			 exclude = gdp->memberlist[i].broadcast;
   exclude++;
   /* data number table init */
   for(i=0;i<MEMBER_COUNT;i++)
   {
	   for(j=0;j<MAX_NUMBER;j++)
		   table[i].number[j] = DATA_NUMBER_INITIALIZED;
	   table[i].index = 0;
   }

   /* mode dependant init */
   switch(s)
   {
   case COLD_INIT : /* cold initialization */
	localdt = DATA_NUMBER_RUNNING_OFFSET;
		localmemberp->status = IDLE;
	full = 0;
	activityproblem = 0;	/* no problem */
	/* applicationtable init */
	for(i=0;i<MAX_APPLICATION_ON_PROCESS;i++)
		{
		reset_application_slot(i);
		reset_application_lock(i);
	}
		/* on exit : delete lock protocol queues, delayed_operation queue */
		atexit(lck_delete);
		atexit(delete_delayed_operation);
	break;
   case COLD_REINIT : /* cold reinitialization */
	localdt = DATA_NUMBER_SILENCE_OFFSET;
	full = 0;
		localmemberp->status = IDLE;
	break;
   case WARM_REINIT : /* warm reinit */
	localdt = DATA_NUMBER_SILENCE_OFFSET;
	full = 1;
	break;
   }

}

/* ker_scheduler()
 * Kernel scheduler.
 *
 * The kernel scheduler manage an internal clock which is able to generate
 * 2 internal time events :
 *      - DECREMENTATION TIME event : when this event is triggered all the
 *        data counters are decremented by period defined in group
 *        Specific routines are called to manage each type of counters:
 *              - lck_validate() lock protocol counter management
 *              - wat_validate_networks() watchdog counter management
 *              - validate_silence() silence period management
 *		- exclude_protocol() exclude command protocol management
 *
 *      - BROADCAST TIME event : when this event is triggered a broadcast
 *        message is sent.
 *
 * The scheduler takes into account external event which are socket event. When
 * a socket event is triggered, a message is read from the socket then :
 *      - the process_message call is activated to run :
 *              - the lock protocol
 *              - the watchdog protocol
 *
 */
int ker_scheduler()
{
int nfound;             /* variable needed for select(2) */
int fire;               /* firewall assertion */

/* DM 14/09/03 replace 'ftime' with 'gettimeofday()' */

//struct timeb scheduler; /* scheduler timeout value */
//struct timeb dcrtime;   /* internal time for next DECREMENTATION_TIME event */
//struct timeb brdtime;   /* internal time for next BROADCAST_TIME event */
//struct timeb value;	/* used to get internal clock value */


struct timeval scheduler_time; /* scheduler timeout value */
struct timeval decrtime;   /* internal time for next DECREMENTATION_TIME event */
struct timeval brdcasttime;   /* internal time for next BROADCAST_TIME event */
struct timeval timeofday;	/* used to get internal clock value */
struct timezone tz;

/**
struct timeb previousvalue;
struct timeb sendtime;
struct timeb previous_sendtime;
**/

/** CDM003 **/
time_t newvalue;      /* used to get internal clock value with local correction
*/
time_t previousvalue; /* used to get save internal clock value */
time_t sendtime;
time_t previous_sendtime;

#define SCHED_SEC scheduler.time
#define SCHED_MSEC scheduler.millitm
#define DCRT_SEC dcrtime.time
#define DCRT_MSEC dcrtime.millitm
#define BRDT_SEC brdtime.time
#define BRDT_MSEC brdtime.millitm
#define VAL_SEC value.time
#define VAL_MSEC value.millitm

#define SCHEDULER_SEC scheduler_time.tv_sec
#define SCHEDULER_USEC scheduler_time.tv_usec
#define DECRT_SEC decrtime.tv_sec
#define DECRT_USEC decrtime.tv_usec
#define BRDCAST_SEC brdcasttime.tv_sec
#define BRDCAST_USEC brdcasttime.tv_usec
#define TIMEOFDAY_SEC timeofday.tv_sec
#define TIMEOFDAY_USEC timeofday.tv_usec


   /* Init */
   status = W_IDLE;
   substatus = SUBIDLE;
   activity_protocol(substatus);
   reset_command_wish(&cmdwish);
   fire = 0;

/* DM 14/09/03 replace 'ftime' with 'gettimeofday()' */
/*	WDX_FTIME(&value); */ /* used to initialize value for activity algorithm */

   gettimeofday (&timeofday, &tz); /* used to initialize value for activity algorithm */

   /* process announce */
   lck_unmark_resource();
   fire = fire | wat_local_member_transition(status, &wdxsend);
   /* outgoing message init */
   memset(mess,'\0',sizeof(struct message));

   /* CDM002 */
   previous_sendtime = (time_t)lib_build_generation();

   lib_forward_announce(mess,localmemberp,build_data_number());
   full = 1;

   /* Infinitive loop */
/* DM 14/09/03 replace 'ftime' with 'gettimeofday()' */
/*   WDX_FTIME(&value); */  /* CDM001*/

	gettimeofday (&timeofday, &tz);

   /* CDM003 - used to initialize value for activity algorithm */
   newvalue = (time_t)lib_build_generation();

   while (!fire)
   {
		/* outgoing message init */
		memset(mess,'\0',sizeof(struct message));
		lck_unmark_resource();
		analyze = 0;

		/*
		 *      Scheduler init : define schedule value
		 *                       define next decrementation_time
		 *                       define next broadcast_time
		 */
/** CDM001      previousvalue = value;
				WDX_FTIME(&value);
				if ( gdp->activity == ENABLE )
					activity_algorithm(&previousvalue,&value,gdp->period);
**/
		/* define next scheduler delay */

/*
		SCHED_MSEC = THOUSAND - VAL_MSEC;
		if ( (SCHED_MSEC == THOUSAND) )
		{
			SCHED_MSEC = 0;
			if ( (VAL_SEC%(gdp->period)) == 0 )
				SCHED_SEC = gdp->period;
			else SCHED_SEC = gdp->period - (VAL_SEC%(gdp->period));
		}
		else
			SCHED_SEC = gdp->period - (VAL_SEC%(gdp->period)) -1;
*/
		SCHEDULER_USEC = MILLION - TIMEOFDAY_USEC;
		if (SCHEDULER_USEC == MILLION)
		{
			SCHEDULER_USEC = 0;
			SCHEDULER_SEC = gdp->period - (TIMEOFDAY_SEC%(gdp->period));
		}
		else
			SCHEDULER_SEC = gdp->period - (TIMEOFDAY_SEC%(gdp->period)) - 1;

		/* define next decrementation time */

/*
		DCRT_SEC = VAL_SEC + (VAL_SEC%(gdp->period));
		if( DCRT_SEC == VAL_SEC )
			DCRT_SEC = VAL_SEC + gdp->period;

		DCRT_MSEC = 0;
*/
		DECRT_SEC = TIMEOFDAY_SEC + (TIMEOFDAY_SEC%(gdp->period));
		if( DECRT_SEC == TIMEOFDAY_SEC )
			DECRT_SEC = TIMEOFDAY_SEC + gdp->period;

		DECRT_USEC = 0;


		/* define next broadcast time */

/*
		BRDT_SEC = VAL_SEC + localbroadcast - ( VAL_SEC%localbroadcast );
		if ( BRDT_SEC == VAL_SEC )
			BRDT_SEC = VAL_SEC + localbroadcast;

		BRDT_MSEC = 0;
*/

		BRDCAST_SEC = TIMEOFDAY_SEC + localbroadcast -
											(TIMEOFDAY_SEC%localbroadcast);
		if (BRDCAST_SEC == TIMEOFDAY_SEC)
			BRDCAST_SEC = TIMEOFDAY_SEC + localbroadcast;

		BRDCAST_USEC = 0;

		wdxsend = 0;
/*
		 nfound = com_select(socketdescriptor,&scheduler);
*/
		nfound = com_select(socketdescriptor,&scheduler_time);

		if ( nfound < 0 )
		{
			 PERROR_B
			 "Can't select event on socket input %m"
			 PERROR_E;
			 fire = fire | -1;
			 continue;
		}

		if ( nfound > 0 )
		{
			 /* read incoming message */
			 input.buffer = (char *) &in_message;
			 input.size = sizeof(struct encoded_message);
#ifndef DEBUG
			 com_get_socket_message(&input);
#endif
			 /* process message */
			 wdxsend = 0;  /* reset the send flag */
			 fire = fire | process_message(&input);
		 if ( fire )
		 return(fire);
		}

/*	previousvalue = value;  */

/* DM 14/09/03 replace 'ftime' with 'gettimeofday()' */
/*        WDX_FTIME(&value); */

		gettimeofday (&timeofday, &tz);

		/* CDM003 */
		previousvalue = newvalue;
		newvalue = (time_t)lib_build_generation();

		if ( gdp->activity == ENABLE )
			activity_algorithm(&previousvalue,&newvalue,gdp->period);

		/*
		 *      DECREMENTATION_TIME EVENT ?
		 */
/*        if ( (VAL_SEC >= DCRT_SEC ) ) */
		if (TIMEOFDAY_SEC >= DECRT_SEC)
		{
			fire = fire | lck_validate(gdp->period,&wdxsend);
			wat_validate_networks(gdp->period);
			analyze++;
			if ( lib_timer_alive(&cmdwish.silence) )
				validate_silence(gdp->period);

			if ( lib_timer_alive(&cmdwish.pause) )
				  validate_pause(gdp->period);

			if ( lib_timer_alive(&cmdwish.exclude) )
				  exclude_protocol(TIMER_EVENT,gdp->period);
		}
		/*
		 *      BROADCAST_TIME EVENT ?
		 */
/*        if ( (VAL_SEC >= BRDT_SEC ) )    */
		if (TIMEOFDAY_SEC >= BRDCAST_SEC)
		{
			 wdxsend++;
		}

		/*
		 *      Member transitions and watchdog protocol
		 */

		if ( analyze )
		{
			wat_threshold();
			wat_analyze_networks();
			wat_remote_member_transition();
			fire = fire | wat_local_member_transition(status, &wdxsend);
			wat_failure(status,&wdxsend);
		}
		/*
		 *      Ready to broadcast message ?
		 */
		if ( wdxsend )
		{
			if ( full )
				 lck_ins_information_in_newlist(&wdxsend);

			/* CDM003 WDX_FTIME( &sendtime ); */
			sendtime = (time_t)lib_build_generation();

			lib_forward_announce(mess,localmemberp,build_data_number());

			/** CDM002 **/
			if ( gdp->activity == ENABLE )
				 broadcast_test(&previous_sendtime, &sendtime);

			previous_sendtime = sendtime;
			/** CDM002 - end **/

			full = 1;
		}
   }
   return(fire);
}

/* void ker_clean_up()
 * Clean up application connection
 *
 * Argument : process pid
 */
void ker_clean_up(p)
pid_t		    p; /* pid of application process */
{
int slot;
int _i;

	slot = NOT_AN_INDEX;
	for(_i = 0; _i < MAX_APPLICATION_ON_PROCESS; _i++)
		if ( locktable[_i].pid == p )
		{
			slot = _i;
			break;
		}

	if ( slot != NOT_AN_INDEX )
	{
		spc_unlock(locktable[slot].lock);
		if ( ( slot = find_application_slot(p)) != NOT_AN_INDEX )
		{
			 com_close_fifo(applicationtable[slot].des);
			 applicationtable[slot].use = SLOT_NOT_USED;
		}
	}
}

/* void ker_build_acknowledge()
 * Build an acknowledge process_message
 * This message is sent to the synchronous FIFO.
 *
 * Argument : process pid
 *	      process status
 *	      identifier
 *	      data
 */
void ker_build_acknowledge(p,s,i,c)
pid_t		    p; /* pid of application process */
enum process_status s; /* status */
int                 i; /* identifier */
char		   *c; /* data */
{
int slot;

	if ( (slot = find_application_slot(p)) != NOT_AN_INDEX )
	{
		if ( send_acknowledge(applicationtable[slot].des,
					s,
						i,
					c) == -1 )
		{
			 com_close_fifo(applicationtable[slot].des);
			 applicationtable[slot].use = SLOT_NOT_USED;
		}
	}
}

/* void ker_build_control()
 * Build and sent a control message
 *
 * Argument : application_slot index
 *	      status
 */
void ker_build_control(i,s)
int i;
enum process_status s;
{
  create_asynchronous_message(i,
							  PROCESS_CONTROL,
							  s,
							  0,
							  lib_build_generation(),
							  NOTKNOWN,
							  "");
}

/* void ker_build_notification()
 * Build and sent a notification message
 *
 * Argument : application process pid
 *	      status
 *	      identifier of the request
 *	      generation number
 *	      resource_status
 *	      data string
 */
void ker_build_notification(p,s,i,g,r,d)
pid_t                p;
enum process_status  s;
int                  i;
int                  g;
enum resource_status r;
char                *d;
{
int slot;

  if ( (slot = find_application_slot(p)) != NOT_AN_INDEX )
  {
		create_asynchronous_message(slot,
									PROCESS_NOTIFICATION,
									s,
									i,
									g,
									r,
									d);
  }
}

/*
 * LOCAL ROUTINES
 */

/* void activity_algorithm()
 * This algorithm is used to detect Operating System scheduling problems. In
 * other words it is used to detect CPU high load.
 *
 * if absolute value of ( current - previous ) is greater than
 * ACTIVITY_CONSTANT wdxperiod
 * then activityproblem is incremented.
 * if activityproblem equals ACTIVITY_TRIGGERING then activity log message is
 * enabled
 *
 * Parameter : pt to previous timer value
 *	       pt to current timer value
 *	       WDX int period
 *
 * Return    : nothing
 */
void activity_algorithm(time_t *p, time_t *c, int e)
{
	if ( (( (long)*c - (long)*p) > ACTIVITY_CONSTANT*e) ||
						(( (long)*p - (long)*c) > ACTIVITY_CONSTANT*e) )
		activityproblem++;
	else
		activityproblem = 0;

	if ( activityproblem == ACTIVITY_TRIGGERING )
	{
		PWARNING_B
		"Activity problem detected by activity algorithm"
		PWARNING_E;
		sprintf(messagestring,"Previous activation was %s",
			  lib_generation_to_s(p));
		PCOMPLEMENT_B
		messagestring
		PCOMPLEMENT_E;
		activityproblem = 0;
	}
}


/* void activity_protocol()
 * Set motus values in the lockprotocol and watchdog linked modules
 * according to the blkprt[] table.
 * Call the update subroutine defined in the blkprt table
 *
 * Argument : substatus
 *
 */
void activity_protocol(s)
enum watchdog_substatus s;
{
struct activity_protocol *pt;      /* pt to activity_protocol table */

	for(pt=actprt;pt->substatus < SUBFAULT;pt++)
	{
	if ( pt->substatus == s )
	{
		if( pt->lockfunct )
		(pt->lockfunct)(pt->lock,pt->script);
		if( pt->watchdogfunct )
		(pt->watchdogfunct)(pt->lock,pt->script);
		break;
	}
	}
}


/* void reset_command_wish()
 * Reset the command_wish data structure
 *
 * Argument : Pt to command_wish data structure
 *
 * Return : None
 */
void reset_command_wish(p)
struct command_wish *p;
{
int _i;

  if(p)
  {
	p->index = NOT_AN_INDEX;
	p->slot = SLOT_NOT_USED;
	memset(&(p->oper),'\0',sizeof(struct operation));
	p->r = RNOREQUEST;
	for(_i=0;_i<MEMBER_COUNT;_i++)
	p->a[_i] = ANOANSWER;
	p->generation = 0;
	lib_timer_reset(&(p->silence));
	lib_timer_reset(&(p->exclude));
	lib_timer_reset(&(p->pause));
	memset(&(p->group1),'\0',sizeof(struct group));
  }
}

/* unsigned int build_data_number()
 * Create next data number value
 *
 * Return : data_number
 */
unsigned int build_data_number()
{
	localdt++;
	return(localdt);
}

/* int validate_data_number()
 * Validate an incoming message
 *
 * Argument : member index
 *            data_number value
 *
 * Return : 0 already received / 1 new data_number
 */
int validate_data_number(index,datanumber)
int index;
unsigned int datanumber;
{
int i;

	for(i=0;i<MAX_NUMBER;i++)
	{
		if (datanumber == table[index].number[i])
			return(0);    /* message already received */
	}
	/* save received datanumber */
	table[index].index = (++table[index].index)%MAX_NUMBER;
	table[index].number[table[index].index] = datanumber;
	return(1);
}

/* void reset_application_slot
 * Reset all the fields of an application_slot structure
 *
 * Argument : slot index
 */
void reset_application_slot(d)
int d;
{
struct application_slot *pt;

   pt = &(applicationtable[d]);
   pt->use = SLOT_NOT_USED;
   pt->pid = 0;
   pt->des = -1;
   pt->mode = NOT_A_MODE;;
   pt->trash = NOT_TRASHED;
   pt->trashgen = 0;
}
void reset_application_lock(i)
int i;
{
struct application_lock *pt;

   pt = &(locktable[i]);
   pt->pid = 0;
   pt->lock = 0;
}

/* int send_acknowledge
 * Build and send an acknowledge message
 *
 * Argument : fifo descriptor
 *	      message status
 *	      message identifier
 *            data pt
 *
 * return :   0 OK / -1 NONOK
 */
int send_acknowledge(d,s,i,c)
int                 d; /* fifo descriptor */
enum process_status s; /* status */
int                 i; /* identifier */
char		   *c; /* data */
{
struct process_acknowledge pra; /* process acknowledge */
struct fifo_output      fo; /* output fifo data structure */

   /* reset buffer */
	memset(&pra,'\0',sizeof(struct process_acknowledge));

	/* set message */
	pra.type = PROCESS_ACKNOWLEDGE;
	pra.status = s;
	pra.identifier = i;
	pra.generation = lib_build_generation();
	strncpy(pra.data,c,PROCESS_DATA_LEN);

	/* send message through asynchronous fifo */
	fo.descriptor = d;
	fo.buffer = (char *) &pra;
	fo.size = sizeof(struct process_acknowledge);
	return( com_send_fifo_s_message(&fo));
}

/* void create_asynchronous_message()
 * Create and send an asychronous message to application process
 * If trash error previously encountered, then a trash message.
 * If trash error encountered, then register this event.
 *
 * If a fatal error occurs during call to this procedure, the slot is closed.
 *
 * Argument : application_slot index
 *	      process type
 *	      process status
 *	      identifier
 *	      generation number
 *            resource status
 *	      data string
 *
 * Return   : None
 */
void create_asynchronous_message(d,y,s,i,g,r,t)
int                 d; /* application slot index */
enum process_type   y; /* message type */
enum process_status s; /* process status */
int                 i; /* request identifier */
int                 g; /* generation number */
int                 r; /* resource status */
char               *t; /* data */
{
struct application_slot *pt; /* pt to application_slot informations */
struct process_message prm;  /* process message */
struct fifo_output      fo;  /* output fifo data structure */
int                     ret; /* call status */
int		       step; /* send message step */
enum process_type        ty; /* process type */
enum process_status      st; /* process_status value */
int			 id; /* identifier */
int 			 ge; /* generation value */
enum resource_status     rs; /* resource status */
char		        *da; /* pt to additional data */

	step = 1;
	pt = &(applicationtable[d]);
	while(step)
	{
	switch(step)
	{
	case 1 : /* normal message send */
			 if ( pt->trash == TRASHED )
			 {
			  /* send TRASHED message */
		ty = PROCESS_CONTROL;
		st = PROCESS_TRASHED;
		id = 0;
		ge = pt->trashgen;
		rs = NOTKNOWN;
		da = NULL;
		step++;
			 }
			 else
			 {
		ty = y;
		st = s;
		id = i;
		ge = g;
		rs = r;
			da = t;
			 }
		 break;
	case 2 : /* trash control send */
		 break;
	}

		/* reset buffer */
		memset(&prm,'\0',sizeof(struct process_message));

		/* set message */
		prm.type = ty;
		prm.status = st;
		prm.identifier = id;
		prm.generation = ge;
	prm.rstatus = rs;
		strncpy(prm.data,da,PROCESS_MESSAGE_LEN);

		/* send message through asynchronous fifo */
		fo.descriptor = pt->des;
		fo.buffer = (char *) &prm;
		fo.size = sizeof(struct process_message);
		ret = com_send_fifo_a_message(&fo);

	switch(ret)
	{
	case 0 : /* OK */
		pt->trash = NOT_TRASHED;
		pt->trashgen = 0;
		step--;
		break;
	case -1 : /* system error */
		step = 0;
		com_close_fifo(pt->des);
		reset_application_slot(d);
		break;
		case -2 : /* PIPE full */
		step = 0;
		pt->trash     = TRASHED;
		pt->trashgen  = lib_build_generation();
		break;
	default :
		break;
	}
	}
}

/* int find_application_slot()
 * Search for the index of a link in the applicationtable
 *
 * Argument : link identifier
 *
 * Return   : index if found
 *            NOT_AN_INDEX if not found
 */
int find_application_slot(id)
int id;
{
int _slot;   /* index in table */
int    _i;   /* index */

   _slot = NOT_AN_INDEX;
   for(_i=0;_i<MAX_APPLICATION_ON_PROCESS;_i++)
	  if( applicationtable[_i].use == SLOT_USED )
		  if ( applicationtable[_i].pid == id )
	  {
		   _slot = _i;
		   break;
	  }
   return(_slot);
}

/* int is_checksum_valid()
 * Validate a message checksum
 *
 * Argument : pointer to encoded_message structure
 *
 * Return   : 0 non valid
 *            1 valid
 */
int is_checksum_valid(em)
struct encoded_message *em;
{
unsigned int lc, lc_rx;

//  Modified by DM 10/10/03
//  put the received and expected checksums into the message

/* control checksum value */
//   lib_char_to_i((WDX_BYTE *)&em->checksum,(WDX_INT *)&lc);

	lib_char_to_i((WDX_BYTE *)&em->checksum,&lc_rx);
	lc = lib_checksum((unsigned char *)&(em->CHECK_MESSAGE),SIZE_MESSAGE);

	if (lc != lc_rx)
	{
		sprintf (messagestring,"Message received with invalid checksum, Recvd=0x%08x, Expctd=0x%08x",lc_rx,lc);
		PWARNING_B
		messagestring
		PWARNING_E;
		return(0);
	}
   return(1);
}

/* int is_message_valid()
 * Validate a message
 * The following fields of the header are checked :
 *	- wdx software version
 *	- mission key
 *	- configuration checksum : after start command, old config version
 *        messages are filtered during silence period.
 *	  otherwise, if a message is received with an invalid configversion
 *	  mission returns to idle state.
 *	- member range
 *
 * Argument : pointer to message structure
 *
 * Return   : 0 non valid
 *            1 valid
 */
int is_message_valid(m)
struct message *m;
{
struct header *h;
char version[10];
char c1;
char c2;
struct c_action a;

   h = &m->head;

/* control wdx version */
   if ( h->wdxversion != gdp->wdxversion )
   {
	sprintf(version,"%o",h->wdxversion);
	sscanf(version,"%c%c",&c1,&c2);
		sprintf(messagestring,
				"Message received with invalid software version (%c.%c)",
		c1,c2);
		PWARNING_B
		messagestring
		PWARNING_E;
		return(0);
   }

/* control missionkey */
   if ( h->missionkey != gdp->missionkey )
		{
		sprintf(messagestring,
			"Message received with invalid mission key (%d)",
		h->missionkey);
		PWARNING_B
	messagestring
		PWARNING_E;
		return(0);
		}

/* control configuration file version */
   if (( h->configversion != gdp->configversion) )
   {
/* configuration checksum : after start command old config version messages are
   filtered during silence period */
		if (( h->configversion == oldversion) && (substatus == SUBSILENCE) &&
		( h->datanumber > DATA_NUMBER_RUNNING_OFFSET ) )
		{
		 PWARNING_B
		 "Message received with previous config version"
		 PWARNING_E;
		}
	else if ( status != W_IDLE )
	{
		 if ( cmdwish.index != NOT_AN_INDEX )
		 {
				  reject_by_prot(&(cmdwish.oper),0);
			  reset_command_wish(&cmdwish);
		 }
		 a.cc = ACSTOP;	   /* action value */
			 strcpy(a.dt,DEFAULT_MISSION);  /* operation data */
			 lib_forward_command(socketdescriptor,
						 TACTION,		/* type */
							 (char *)&a,
						 build_data_number());
		 sprintf(messagestring,
			 "Message received with invalid configuration file version (%u)",
		 h->configversion);
			 PWARNING_B
		 messagestring
			 PWARNING_E;
	}
		return(0);
   }

/* control memberindex range */
   if ( (h->memberindex <= NOT_AN_INDEX) || (h->memberindex >= gdp->membercount) )
   {
	sprintf(messagestring,
				"Message received with invalid member index (%d)",
		h->memberindex);
		PWARNING_B
	messagestring
		PWARNING_E;
		return(0);
   }

  return(1);
}

/* validate_silence()
 * Decrements the silence counter of the command_wish data structure
 * If timer event is raised then CTIMER event is triggered.
 *
 * Argument : decrementation value
 *
 * Return : None
 */
void validate_silence(t)
int t;	    /* decrementation value */
{
struct c_action a;  /* command_action structure */

	lib_timer_decrease(&(cmdwish.silence),t);
	if ( lib_timer_event(&(cmdwish.silence)) )
	{
	   lib_timer_reset(&(cmdwish.silence));
	   a.cc = ACTIMER;			/* ACTIMER action */
	   lib_forward_command(socketdescriptor,
			   TACTION,		/* type */
			   (char *)&a,
			   build_data_number());
	}
}

/* validate_pause()
 * Decrements the pause counter of the command_wish data structure
 * If timer event is raised then CPAUSETIMER event is triggered.
 *
 * Argument : decrementation value
 *
 * Return : None
 */
void validate_pause(t)
int t;	    /* decrementation value */
{
struct c_action a;  /* command_action structure */

	lib_timer_decrease(&(cmdwish.pause),t);
	if ( lib_timer_event(&(cmdwish.pause)) )
	{
	   lib_timer_reset(&(cmdwish.pause));
	   a.cc = ACPAUSETIMER;			/* ACPAUSETIMER action */
	   lib_forward_command(socketdescriptor,
			   TACTION,		/* type */
			   (char *)&a,
			   build_data_number());
	}
}

/* enum command_answer answer_to_start_request()
 * A checksum has been received. Read the configuration files, store the
 * result into the cmdwish.group1 structure and return AOK if checksum is OK
 * ANONOK otherwise.
 *
 * BE CAREFULL : checksum are generated as unsigned but transmitted as integer
 *
 * Argument : received checksum
 *
 * Return : AOK or ANONOK
 */
enum command_answer answer_to_start_request(c)
int c;
{
int b;
/*
	if(cmdwish.group1 == NULL)
		 cmdwish.group1 = (struct group *)  malloc(sizeof(struct group));
	if(cmdwish.group1 == NULL)
	{
	PERROR_B
	"Memory allocation for structure group %m"
	PERROR_E;
	return(ANONOK);
	}
*/
	if ( cfg_config_init(gdp->mission,&cmdwish.group1) == 0 )
	{
	b = (int) cmdwish.group1.configversion;
	if ( b == c )
		return(AOK);
	}
	return(ANONOK);
}

/* void exclude_protocol()
 * Performs the exclude protocol
 * Look to the answer table of the command_wish element to determine the
 * status of a request.
 * If event is TIME, timer is decreased by t
 * If all the answer are OK or timeout reached ( end of protocol ), then send
 * an ACTION over the network.
 *
 * Argument : event : NETWORK or TIME
 *
 * Return   : None
 */
void exclude_protocol(e,t)
enum protocol_event e; /* event type */
int                 t; /* decrementation value */
{
int                    _i; /* index */
int                _index; /* number of OK or timeout */
int               _reject; /* number of NONOK */
struct c_action         a; /* action message to send */
struct request_action *pt; /* pt to conversion table */

   _reject = 0;
   _index = 0;

   /* on event */
   switch(e)
   {
   case NETWORK_EVENT :
	break;
   case TIMER_EVENT :
	lib_timer_decrease(&(cmdwish.exclude),t);
	break;
   }

   /* look at the answer table */
   for( _i=0;_i < gdp->membercount; _i++)
		{
			 switch(cmdwish.a[_i])
				  {
				  case ANOANSWER :
					   if ( lib_timer_event(&(cmdwish.exclude)) )
							_index++;
					   else {
							if ( !lib_is_member_valid(gdp->memberlist,_i) )
								 _index++;
							}
					   break;
				  case AOK :
					   _index++;
					   break;
				  case ANONOK :
			   _reject++;
					   break;
				  default :
			   _reject++;
					   break;
				  }
		}
	if ( _reject != 0 )
	{
	/* reject */
	reject_by_prot(&(cmdwish.oper),0);
	reset_command_wish(&cmdwish);
	}
	else if ( _index == gdp->membercount )
	{
	   /* action */
	   for(pt=reqact;pt->value;pt++)
		if ( pt->value == cmdwish.r )
		{
		a.cc = pt->action;		   /* action value */
		break;
		}
	   a.a1 = cmdwish.oper.operationargument;    /* operation argument */
	   strcpy(a.dt,cmdwish.oper.operationdata);  /* operation data */
	   lib_forward_command(socketdescriptor,
			   TACTION,		/* type */
			   (char *)&a,
			   build_data_number());
	   /* reset timer */
	   lib_timer_reset(&(cmdwish.exclude));
	}
}

/* int exclude_act()
 * Receive an answer. If the request is local, then active the exclusion
 * protocol.
 *
 * Argument : pointer to command
 *	      answer received
 *
 * Return   : 0 ok or -1 fatal error
 */
int exclude_act(c,a)
struct command     *c; /* command pt */
enum command_answer a; /* answer received value */
{
int               ret; /* returned value */
struct c_answer    *w; /* c_answer pt */

   /* success init */
   ret = 0;

   if(c)
   {
	   w = (struct c_answer *)&(c->value1);
	   /* check only for local request, and exclude wrong answers */
	   if ( ( w->ri == localindex ) &&
		( w->cr == cmdwish.r ) &&
		( w->gn == cmdwish.generation ) )
	   {
		cmdwish.a[c->index] = a;   /* store answer */
		exclude_protocol(NETWORK_EVENT,0); /* active the exclude protocol */
	   }
   }
   return(0);
}

/* int ignore_act()
 * Ignore the incoming action
 *
 * Argument : pointer to command
 *	      action
 *
 * Return   : 0 ok or -1 fatal error
 */
int ignore_act(c,a)
struct command     *c; /* command pt */
enum command_action a; /* action value */
{
   return(0);
}

/* int conflict_act()
 * A conflict exist between a remote request and a local request :
 * The older requets is saved.
 * If the local request is the older : forward a ANONOK answer message.
 * If the remote request is the older, then erase the local request
 * and forward a AOK answer message.
 *
 * Argument : pointer to command
 *	      command request received
 *
 * Return   : 0 ok : status updated or -1 : status not updated
 */
int conflict_act(c,t)
struct command      *c; /* command pt */
enum command_request t; /* request received */
{
int               ret; /* returned value */
struct c_request   *r; /* c_request pt */
enum command_answer w; /* answer */
struct c_answer     a; /* answer data structure */

   /* success init */
   ret = 0;

   if(c)
   {
	   r = (struct c_request *)&(c->value1);
	   if ( c->index != localindex ) /* check only remote request */
	   {
		if ( r->gn < cmdwish.generation )
		{
		/* reject local request */
			reject_by_prot(&(cmdwish.oper),0);
		reset_command_wish(&cmdwish);
		/* check request */
		/* answer */
				if ( r->cr == RSTART )
					 w = answer_to_start_request((unsigned int)r->ar);
				if ( r->cr == RPAUSESTART )
					 w = answer_to_start_request((unsigned int)r->ar);
				else w = AOK;
		ret = 0;    /* status updated */
		}
		else /* local is the older */
		{
		w = ANONOK;
		ret = 1;    /* status not updated */
		}
	   }
	   else
	   {
			w  = AOK; /* send OK to myself */
		ret = 1;  /* status not updated */
	   }
	   a.ri = c->index; /* requester index */
	   a.cr = r->cr;	/* request */
	   a.gn = r->gn;    /* request generation */
	   a.ca = w;	/* answer */
	   lib_forward_command(socketdescriptor,
			   TANSWER,	    /* type */
			   (char *)&a,
			   build_data_number());
   }
   return(ret);
}

/* int reject_act()
 * Send a NONOK answer command message
 * Reject a command request
 *
 * Argument : pointer to command
 *	      command request : not used
 *
 * Return   : 0 ok or -1 fatal error
 */
int reject_act(c,t)
struct command       *c; /* command pt */
enum command_request  t; /* command request : NOT USED */
{
int               ret; /* returned value */
struct c_answer     w; /* answer structure */
struct c_request   *r; /* pt to request structure */

   /* success init */
   ret = 0;

   /* if command pointer valid : forward answer */
   if(c)
   {
	   r = (struct c_request *) &(c->value1);
	   w.ri = c->index; /* requester index */
	   w.cr = r->cr;    /* request */
	   w.gn = r->gn;    /* generation number of the request */
	   w.ca = ANONOK;	/* answer */
	   lib_forward_command(socketdescriptor,
			   TANSWER,
			   (char *)&w,
			   build_data_number());
   }
   return(0);
}

/* int accept_act()
 * Send an AOK answer command message. A request has been received. Create my
 * answer.
 * For RSTART request, read the config files then check the configversion with
 * the value received. If identical send AOK else send ANONOK
 *
 * Argument : pointer to command
 *	      command request received
 *
 * Return   : 0 ok or -1 fatal error
 */
int accept_act(c,a)
struct command      *c; /* command pt */
enum command_request a; /* request value */
{
int               ret; /* returned value */
struct c_answer     w; /* answer structure */
struct c_request   *r; /* pt to request structure */

   /* success init */
   ret = 0;

   /* if command pointer valid : forward answer */
   if(c)
   {
	   r = (struct c_request *) &(c->value1);
	   w.ri = c->index; /* requester index */
	   w.cr = r->cr;    /* request */
	   w.gn = r->gn;    /* generation number of the request */
	   /* answer */
	   if ( r->cr == RSTART || r->cr == RPAUSESTART )
			w.ca = answer_to_start_request((unsigned int)r->ar);
	   else w.ca = AOK;
	   lib_forward_command(socketdescriptor,
			   TANSWER,
			   (char *)&w,
			   build_data_number());
   }
   return(0);
}

/* int process_act()
 * Process on action message : an action is received. process it
 *
 * Argument : pointer to command
 *	      parameter command_action
 *
 * Return   : 0 ok or -1 fatal error
 */
int process_act(c,t)
struct command   *c; /* command pt */
enum command_action t; /* action type */
{
char                   *s; /* pointer to message */
int                   ret; /* returned value */
int		    reset; /* reset command wish if set */
int		        i; /* index */
enum process_status  stat; /* message status */
struct c_action        *a; /* pt to c_action data */
struct delayed_operation *_dc; /* pt to delayed operation structure */
struct delayed_operation *_dn; /* pt to delayed operation structure */
int oldindex;
struct c_request        r;  /* command request data structure */

   /* success init */
   ret = 0;
   reset = 0;
   stat = PROCESS_SUCCESS;

   oldindex = localindex;   /* saved in case of RSTART command */
   if(c)
   {
	a = ( struct c_action *) &(c->value1);
		switch(t)
		{
	case ACTIMER : /* timeout end action */
		/* the delayed lock requests are processed */
		if ( c->index == localindex )
		{
		 localdt = DATA_NUMBER_RUNNING_OFFSET; /* datanumber resetted */
				 _dc = delayedroot;
				 while(_dc)
				 {
					   lock_opr(&_dc->oper,_dc->request);
			   _dn = _dc;
					   _dc = _dc->next;
			   free(_dn);
		 }
		 delayedroot = NULL;
		 lck_reset_resource();   /* reset resources with no owner */
		 reset_command_wish(&cmdwish); /* reset command wish data */
		}
		else ret = 1;
		return(ret);
		break;
	case ACUP : /* MEMBER UP action */
		if ( c->index == localindex )
		{
/* lionel tricon */
				 s = "Local member is running"; /* message */
			 ret = lib_process_init(COLD_REINIT_PROCESS,
						NULL,
						gdp,
						NULL,
						mess);	/* cold startup */
			 if ( ret == -1 )
			 {	/* exit */
			 status = W_FAULT;
			 substatus = SUBFAULT;
			 return(1);
			 }
		 lib_timer_set(&(cmdwish.silence),silence);
		 wdxsend++;		        /* broadcast message */
		}
		else return(1);
		break;
		case ACRUN : /* RUN action */
		s = "All members are running";	/* message */
		ret = lib_process_init(COLD_REINIT_PROCESS,
				   NULL,
				   gdp,
				   NULL,
				   mess);	/* cold startup */
		if ( ret == -1 )
		{	/* exit */
		status = W_FAULT;
		substatus = SUBFAULT;
		return(1);
		}
		lib_timer_set(&(cmdwish.silence),silence);
		wdxsend++;			        /* broadcast message */
		break;
		case ACPAUSESTART : /* START after a Pause */
			/* There is no Break for this case because we want just */
			/* to do as for a ACSTART but having no index defined */
			/* to avoid a bad error message */

			c->index = NOT_AN_INDEX;

		case ACSTART : /* START action */
		s = "Mission started";		/* message */

		/* init watchdog with saved snapshot */
		ret = lib_process_init(WARM_REINIT_PROCESS,
				   c->commanddata,
				   gdp,
				   &cmdwish.group1,
				   mess);	/* warm_startup */
		if ( ret == -1 )
		{	/* exit */
		status = W_FAULT;
		substatus = SUBFAULT;
		return(1);
		}
		/* setup silence timer */
		lib_timer_set(&(cmdwish.silence),silence);
		analyze++;	/* force member status analyze */
		wdxsend++;	/* broadcast message */
		break;
		case ACSTOP : /* STOP action */
		reset++;				/* reset cmdwish */
		lib_set_snapshot(c->commanddata,gdp);
		/* activate transition to idle status */
		for(i=0;i<gdp->membercount;i++)
		if(gdp->memberlist[i].status != FAULTY )
		   gdp->memberlist[i].brdstatus = IDLE;
		wat_remote_member_transition();
			wat_local_member_transition(W_IDLE, &wdxsend);
		s = "Mission stopped, all members are IDLE";		/* message */
		break;

		case ACPAUSE : /* PAUSE action */
		lib_set_snapshot(c->commanddata,gdp);
		/* activate transition to idle status */
		for(i=0;i<gdp->membercount;i++)
		if(gdp->memberlist[i].status != FAULTY )
		   gdp->memberlist[i].brdstatus = IDLE;
		wat_remote_member_transition();
			wat_local_member_transition(W_IDLE, &wdxsend);
		s = "Mission paused, all members are IDLE";		/* message */
			lib_timer_set(&(cmdwish.pause),a->a1);
		break;

		case ACPAUSETIMER :  /* 	Pause timeout */
		if ( c->index == localindex )
			{
				reset_command_wish(&cmdwish); /* reset command wish data */
				if(cfg_config_init(gdp->mission,&cmdwish.group1) != -1 )
				{
					 cmdwish.oper.index = localindex;
					 cmdwish.oper.pid = NOT_A_PID;
					 cmdwish.oper.type = RPAUSESTART;
					 /* the mission name is forced to the default value:
						mission name */
					  strcpy(cmdwish.oper.operationdata,gdp->mission);
			 cmdwish.r = RPAUSESTART;
					 cmdwish.generation = lib_build_generation();
			 lib_timer_set(&(cmdwish.exclude),exclude);
					 r.cr = cmdwish.oper.type;
					 r.gn = cmdwish.generation;
					 r.ar = (int) cmdwish.group1.configversion;
					 lib_forward_command(socketdescriptor,
									   TREQUEST,
						   (char *)&r,
							   build_data_number());
				 }
			 }
			 else ret = 1;
			 return (ret);
			 break;


		case ACSHUT : /* EXCEPTION action */
		if ( c->index == localindex )
			 reset++;			/* reset cmdwish */
			s = "Exception script triggered";	/* message */
			wat_node_exception(OPERATOR,s);	/* trigger exception script */
		break;
		case ACLOCKILL : /* PROC KILL (leave) action */
		if ( c->index == localindex )	/* only if local command */
		{
			 reset++;			/* reset cmdwish */
/* lionel tricon */
				 s = "Local wdx process killed";/* message */
		}
		else return(1);
			break;
		case ACKILL : /* KILL action */
		reset++;				/* reset cmdwish */
			s = "All wdx processes in mission killed"; /* message */
			break;
		default :  /* other : bugg */
		sprintf(messagestring,"Process_act(), invalid parameter %d",t);
		s = messagestring;
		stat = PROCESS_WARNING;
		break;
		}
		/* send message to log file */
		switch(stat)
		{
		case PROCESS_SUCCESS :
		PINFOR_B
		s
		PINFOR_E;
		break;
		case PROCESS_ERROR :
		PERROR_B
		s
		PERROR_E;
		break;
		case PROCESS_WARNING :
		PWARNING_B
		s
		PWARNING_E;
		break;
		}
		/* search for slot for local request only */

		if ( ( c->index == oldindex ) )
		{
			ker_build_acknowledge(cmdwish.oper.pid,
							   PROCESS_SUCCESS,
					   cmdwish.oper.identifier,
					   s);
		if(reset)	/* reset cmdwish if flag set */
			reset_command_wish(&cmdwish);
		}
   }
   return(ret);
}

/* int reject_by_prot()
 * Exclude protocol reject an operation message : send to the calling
 * application a reject message.
 *
 * Argument : pointer to operation
 *	      parameter ( NOT used : message type to forward )
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int reject_by_prot(o,t)
struct operation *o; /* operation pt */
int               t; /* request type */
{
char                   *s; /* pointer to message */
int                   ret; /* returned value */

   /* success init */
   ret = 0;

   /* search for slot for local request only */
   s = "Command rejected by mission";
   ker_build_acknowledge(o->pid,
					 PROCESS_ERROR,
			 o->identifier,
			 s);
   return(ret);
}

/* int aldone_opr()
 * An action is already done. Return OK message
 *
 * Argument : pointer to operation
 *	      action value
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int aldone_opr(o,a)
struct operation   *o; /* operation pt */
enum command_action a; /* action value */
{
char                   *s; /* pointer to message */
int                   ret; /* returned value */
enum process_status  stat; /* message status */

   /* success init */
   ret = 0;
   stat = PROCESS_SUCCESS;

   if(o)
   {
		switch(a)
		{
	case ACTIMER : /* timeout end action */
		break;
	case ACUP : /* MEMBER UP action */
		if ( o->index == localindex )
		{
/* lionel tricon */
				 s = "Local member is running"; /* message */
		}
		else return(1);
		break;
		case ACRUN : /* RUN action */
		s = "All members are running";	/* message */
		break;
		case ACSTART : /* START action */
		s = "Mission started";		/* message */
		break;
		case ACSTOP : /* STOP action */
		s = "Mission stopped";		/* message */
		break;
		case ACPAUSE : /* PAUSE action */
		s = "Mission paused ";		/* message */
		break;
		case ACPAUSETIMER : /* TIMER PAUSE action */
		s = "Pause timer running ";		/* message */
		break;
		case ACPAUSESTART : /* TIMERPAUSE expired */
		s = "Pause timer terminated";		/* message */
		break;
		case ACSHUT : /* EXCEPTION action */
			s = "Exception script triggered";	/* message */
		break;
		case ACLOCKILL : /* PROC KILL action */
		if ( o->index == localindex )	/* only if local command */
		{
				 s = "Local wdx process killed";	/* message */
		}
		else return(1);
			break;
		case ACKILL : /* KILL action */
			s = "All wdx processes in mission killed"; /* message */
			break;
		default :  /* other : bugg */
		sprintf(messagestring,"Aldone_opr(), invalid parameter %d",a);
		s = messagestring;
		stat = PROCESS_WARNING;
		break;
		}
		/* send message to log file */
		switch(stat)
		{
		case PROCESS_SUCCESS :
		PINFOR_B
		s
		PINFOR_E;
		break;
		case PROCESS_ERROR :
		PERROR_B
		s
		PERROR_E;
		break;
		case PROCESS_WARNING :
		PWARNING_B
		s
		PWARNING_E;
		break;
		}
		/* search for slot for local request only */
		if ( ( o->index == localindex ) )
		{
			ker_build_acknowledge(o->pid,
								   PROCESS_SUCCESS,
						   o->identifier,
						   s);
		}
   }
   return(ret);
}

/* int action_opr()
 * Send an action command message : set the command wish data structure then
 * broadcast an action command messsage over the network.
 *
 * Argument : pointer to operation
 *	      action value to send
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int action_opr(o,a)
struct operation   *o; /* operation pt */
enum command_action a; /* action value */
{
int               ret; /* returned value */
struct c_action     t; /* action structure */
int              slot; /* slot value */


   /* success init */
   ret = 0;

   if ( (slot = find_application_slot(o->pid)) != NOT_AN_INDEX  && o )
   {
	   if ( cmdwish.index != NOT_AN_INDEX )
	   {
		reject_by_prot(&(cmdwish.oper),0);
		reset_command_wish(&cmdwish);
	   }
	   cmdwish.index = localindex;
	   cmdwish.slot = slot;
	   cmdwish.oper = *o;	/* save operation */
	   cmdwish.c = a;		/* save action */
	   cmdwish.generation = lib_build_generation();
	   t.cc = a;    /* action value */
	   t.a1 = o->operationargument;
	   strcpy(t.dt,o->operationdata);
	   lib_forward_command(socketdescriptor,
			   TACTION,
			   (char *)&t,
			   build_data_number());
   }
   return(0);
}

/* int reject_opr()
 * Reject an operation message : send to the calling application a reject
 * message.
 *
 * Argument : pointer to operation
 *	      parameter ( NOT used : message type to forward )
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int reject_opr(o,t)
struct operation *o; /* operation pt */
int               t; /* request type */
{
char                   *s; /* pointer to message */
int                   ret; /* returned value */

   /* success init */
   ret = 0;

   /* search for slot for local request only */
   s = "Command not allowed : bad process status";
   ker_build_acknowledge(o->pid,
					 PROCESS_ERROR,
			 o->identifier,
			 s);
   return(ret);
}

/* int request_opr()
 * Process a request operation message
 * Set the command_wish data structure then broadcast the request over the
 * network.
 *
 * BE CAREFULL : when configversion is forwarded over the network a cast
 * from unsigned int to int is performed.
 *
 * Argument : pointer to operation
 *	      message type to forward
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int request_opr(o,t)
struct operation *o; /* operation pt */
int               t; /* request type */
{
int                  slot; /* index in applicationtable */
char                   *s; /* pointer to message */
int                   ret; /* returned value */
struct c_request        r; /* request structure */


   /* success init */
   ret = 0;

   /* search for slot for local request only */
   if ( ( o->index == localindex ) &&
		( (slot = find_application_slot(o->pid)) != NOT_AN_INDEX ) )
   {
	   cmdwish.index = localindex;
	   cmdwish.slot = slot;
	   cmdwish.oper = *o;	/* save operation */
	   cmdwish.r = t;
	   cmdwish.generation = lib_build_generation();
	   lib_timer_set(&(cmdwish.exclude),exclude);
	   r.cr = cmdwish.r;
	   r.gn = cmdwish.generation;
	   r.ar = 0;

	   switch(t)
	   {
	   case RSTART :	/* start */
		if(cfg_config_init(gdp->mission,&cmdwish.group1) == -1 )
		{
		s = "Configuration mission.cfg file error";
		ker_build_acknowledge(o->pid,
							  PROCESS_ERROR,
					  o->identifier,
					  s);
		reset_command_wish(&cmdwish);
		return(-1);
		}
		r.ar = (int) cmdwish.group1.configversion;
		break;
	   default :	/* other */
		break;
	   }
	   lib_forward_command(socketdescriptor,
			   TREQUEST,
			   (char *)&r,
			   build_data_number());
   }
   return(ret);
}

/* int local_opr()
 * Process a local operation message : format output data and send them to
 * the application
 *
 * Argument : pointer to operation
 *	      operation value
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int local_opr(o,t)
struct operation *o;  /* operation */
int               t;  /* operation type */
{
enum process_status  stat; /* acknowledge status */
char                   *s; /* pointer to message */
int                   ret; /* returned value */
char  d[PROCESS_DATA_LEN]; /* data formatted string */
char  format1[STRING_LEN]; /* output format */
int                     i; /* index */
int                     j; /* index */
int                   own; /* master token owner index */
enum resource_status status; /* status of mastertoken */


   /* success init */
   ret = 0;
   stat = PROCESS_SUCCESS;
   /* init of d */
   d[0] = '\0';

   switch(t)
   {
   case WDX_LOG : /* log operation : local */
	PINFOR_B
		"Log file closed by operator"
	PINFOR_E;
		lib_redirect_output(gdp->log,gdp->logfile);
		PINFOR_B
	"Log file opened by operator"
	PINFOR_E;
	s = "Log file initialized";
		break;
   case WDX_CONSTANT : /* version operation : local */
		sprintf(d,"WDX configuration parameters for mission \"%s\":\n",gdp->mission);
	sprintf(d+strlen(d),
				"(values marked with (*) are configurable in mission and local files)\n");
		sprintf(d+strlen(d),
				"Software version                  = %s\n",WDX_VERSION_STRING);
		sprintf(d+strlen(d),
				"Configuration file(*)             = V%d.%d - %s\n",
				(int)gdp->version[0],
				(int)gdp->version[1],
		gdp->date);
		sprintf(d+strlen(d),
				"Configuration checksum            = %u\n",gdp->configversion);
		sprintf(d+strlen(d),
				"Scheduler period                  = %d second(s)\n",gdp->period);
		sprintf(d+strlen(d),
				"Process priority                  = %d\n",gdp->priority);
	sprintf(d+strlen(d),
				"Member name(*)                    = %s\n",localmemberp->name);
	sprintf(d+strlen(d),
				"Member type (*)                   = %s\n",
		(localmemberp->type == MASTER_SUBSTITUTE_T) ? PRINTABLE_MAS_SUB : PRINTABLE_SAT);
	sprintf(d+strlen(d),
				"MASTERTOKEN priority(*)           = %d\n",localmemberp->priority);
		sprintf(d+strlen(d),
				"Broadcast time(*)                 = %d\n",localmemberp->broadcast);
	sprintf(d+strlen(d),
				"Validity time(*)                  = %d\n",localmemberp->validity);
	sprintf(d+strlen(d),
				"Redundancy(*)                     = %d\n",
			localmemberp->redundancy);
		for ( i=0;i<NETWORKS;i++)
			   sprintf(d+strlen(d),
				"Interface %d(*)                    = %s\n",
					  i+1,localmemberp->networks[i].ifname);
		sprintf(d+strlen(d),
				"UDP service name                  = %s",
			com_get_service_name(socketdescriptor));
		s = d;
	break;
   case WDX_SCRIPT : /* constant operation : second part */
		sprintf(d+strlen(d),
				"NEWMASTERSCRIPT(*)                = %s\n",gdp->newmasterscript);
		sprintf(d+strlen(d),
				"NOSUBSTITUTESCRIPT(*)             = %s\n",gdp->nosubstitutescript);
		sprintf(d+strlen(d),
				"SATELLITESCRIPT(*)                = %s\n",gdp->satellitescript);
		sprintf(d+strlen(d),
				"EXCEPTIONSCRIPT(*)                = %s\n",gdp->exceptionscript);
		sprintf(d+strlen(d),
				"EVENTSCRIPT(*)                    = %s\n",gdp->eventscript);
		for(i=0;i<MEMBER_STATUS;i++)
		for(j=0;j<MEMBER_STATUS;j++)
	{
		if ( i != j )
				sprintf(d+strlen(d),
					"%-30.30s(*) = %s\n",
			gdp->membermasks[i][j].key,
			(gdp->membermasks[i][j].value == YES) ? PRINTABLE_YES : PRINTABLE_NO);
		}
		for(i=UNKNOWN;i<NETWORK_STATUS;i++)
			for(j=UNKNOWN;j<NETWORK_STATUS;j++)
		{
		if ( i != j )
				sprintf(d+strlen(d),
					"%-30.30s(*) = %s\n",
			gdp->networkmasks[i][j].key,
			(gdp->networkmasks[i][j].value == YES) ? PRINTABLE_YES : PRINTABLE_NO);
		}
		for(i=0;i<RESOURCE_STATUS;i++)
		for(j=0;j<RESOURCE_STATUS;j++)
				sprintf(d+strlen(d),
					"%-30.30s(*) = %s\n",
			gdp->resourcemasks[i][j].key,
			(gdp->resourcemasks[i][j].value == YES) ? PRINTABLE_YES : PRINTABLE_NO);
		sprintf(d+strlen(d),
				"%-30.30s(*) = %s\n",
		gdp->mismatchmask.key,
		(gdp->mismatchmask.value == YES) ? PRINTABLE_YES : PRINTABLE_NO);
		s = d;
	break;
   case WDX_STATUS : /* status operation : local */
	sprintf(d,"%s",member_status_to_s(localmemberp->status));
	s = d;
	break;
   case WDX_LOCALNAME : /* local name operation */
	sprintf(d,"%s",localmemberp->name);
	s = d;
	break;
   case WDX_MASTERNAME : /* mastername operation : local */
		if ( (status = lib_status_of_resource(gdp->resourcelist,
											  MASTER_TOKEN)) == RESERVED )
		{
			 own = lib_requester_of_resource(gdp->resourcelist,MASTER_TOKEN);
			 sprintf(d,"%s",gdp->memberlist[own].name);
		}
		else sprintf(d,MASTERHOLDER_IS_UNKNOWN);
	s = d;
	break;
   case WDX_LIST : /* list operation : local */
		sprintf(d,"\nStatus of mission \"%s\" on member \"%s\"\n\n",
				gdp->mission,localmemberp->name);
		sprintf(format1,
				"%%-%ds  %%-%ds  %%-%ds  %%-%ds\n",
				MEMBER_FORMAT,
				MEMBER_STATUS_FORMAT,
				NETWORK_FORMAT,
				NETWORK_FORMAT);

		sprintf(d+strlen(d),format1,
				"MEMBER","STATUS","NETWORK-1","NETWORK-2");
		for (i=0; i < gdp->membercount; i++)
			 sprintf(d+strlen(d),format1,
					 gdp->memberlist[i].name,
					 member_status_to_s(gdp->memberlist[i].status),
		network_status_to_s(gdp->memberlist[i].networks[NETWORK_1].status),
		network_status_to_s(gdp->memberlist[i].networks[NETWORK_2].status));
		sprintf(d+strlen(d),"\n");
		sprintf(format1,
				"%%-%ds  %%-%ds  %%-%ds  %%-%ds\n",
				RESOURCE_FORMAT,
				RESOURCE_STATUS_FORMAT,
				OWNER_FORMAT,
				GENERATION_FORMAT);

		sprintf(d+strlen(d),format1,
			"RESOURCE/DIRECTORY","STATUS","BY","AT");
		for (i=MASTER_INDEX; i < RESOURCE_COUNT; i++)
		{
			 if ( i == MASTER_INDEX )
				  s = MASTER_TOKEN_STRING;
			 else s = gdp->resourcelist[i].name;
			 if ( gdp->resourcelist[i].status != NOTKNOWN )
			 {
				  if ( gdp->resourcelist[i].initiator == NOT_AN_OWNER )
					   sprintf(d+strlen(d),format1,
					   s,
					   resource_status_to_s(gdp->resourcelist[i].status),
					   "","");
				  else sprintf(d+strlen(d),format1,
						  s,
						  resource_status_to_s(gdp->resourcelist[i].status),
						  gdp->memberlist[gdp->resourcelist[i].initiator].name,
						 lib_generation_to_s(&gdp->resourcelist[i].generation));
			 }
		}
	s = d;
	break;
   default : /* value not supported */
		sprintf(messagestring,
				"local_opr(), invalid local operation %d",
				t);
	stat = PROCESS_WARNING;
	s = "Invalid local operation parameter";
	PWARNING_B
	s
	PWARNING_E;
		break;
   }
   /* send acknowledge for local application only */
   ker_build_acknowledge(o->pid,
					 stat,
			 o->identifier,
			 s);
   return(ret);
}

/* int wait_opr()
 * Process a wait operation message : call the lock protocol
 *
 * Argument : pointer to operation
 *	      operation value
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int wait_opr(c,t)
struct operation *c;  /* operation */
int               t;  /* operation type */
{
int  ret; /* returned value */
char *s; /* pt to message string */


   ret = lck_wait_operation_input(c,t);
   if ( ret == 0 )
		s = "Command processing";
   else s = "Command rejected by wdx lock protocol";
   ker_build_acknowledge(c->pid,PROCESS_SUCCESS,c->identifier,s);
   return(0);
}

/* int master_opr()
 * Process a master operation message : call the lock_opr routine
 * with MASTERTOKEN as resource value
 * Argument : pointer to operation
 *	      operation value
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int master_opr(c,t)
struct operation *c;  /* operation */
enum lock_request t;  /* operation type */
{
char *s;
int  lock_stat;


   if ( localmemberp->type != MASTER_SUBSTITUTE_T )
   {
	s = "Command not allowed for satellite";
	ker_build_acknowledge(c->pid,
				  PROCESS_ERROR,
				  c->identifier,
				  s);
		return(0);
   }
   else
   {
		strcpy(c->operationdata,MASTER_TOKEN);
		lock_stat = lock_opr(c,t);
		return(lock_stat);

   }
}

/* int masdel_opr()
 * Delay a master operation message : call delay_opr routine with
 * Mastertoken as resource value
 *
 * Argument : pointer to operation
 *	      operation value
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int masdel_opr(c,t)
struct operation *c;  /* operation */
enum lock_request t;  /* operation type */
{
char *s;


   if ( localmemberp->type != MASTER_SUBSTITUTE_T )
   {
	s = "Command not allowed for satellite";
	ker_build_acknowledge(c->pid,
				  PROCESS_ERROR,
				  c->identifier,
				  s);
		return(0);
   }
   else
   {
		strcpy(c->operationdata,MASTER_TOKEN);
		return(delay_opr(c,t));
   }
}
/* int masidl_opr()
 * Initiate a setup action, then call the masdel_opr() routine.
 *
 * Argument : pointer to operation
 *	      operation value
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int masidl_opr(c,t)
struct operation *c;  /* operation */
enum lock_request t;  /* operation type */
{
char *s;
struct c_action a;


   if ( localmemberp->type != MASTER_SUBSTITUTE_T )
   {
	s = "Command not allowed for satellite";
	ker_build_acknowledge(c->pid,
				  PROCESS_ERROR,
				  c->identifier,
				  s);
		return(0);
   }
   else
   {
		a.cc = ACUP;    /* action value */
		a.a1 = 0;
		a.dt[0] = '\0';
		lib_forward_command(socketdescriptor,
			   TACTION,
			   (char *)&a,
			   build_data_number());
		strcpy(c->operationdata,MASTER_TOKEN);
		return(delay_opr(c,t));
   }
}

/* int delay_opr()
 * Delay a lock operation message using the delay operation queue
 *
 * Argument : pointer to operation
 *	      operation value
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int delay_opr(c,t)
struct operation *c;  /* operation */
enum lock_request t;  /* operation type */
{
int                   ret; /* returned value */
char		       *s; /* pt to message string */
struct delayed_operation *_dc;  /* delayed operation structure pt */
struct delayed_operation *_dcc; /* delayed operation structure pt */


   /* success init */
   ret = 0;

   /* silence : operation is delayed */
   _dc = (struct delayed_operation *) malloc(sizeof(struct delayed_operation));
   if (!_dc)
   {
	   s = "Can't malloc for delayed_operation data structure";
	   PERROR_B
	   s
	   PERROR_E;
	   ker_build_acknowledge(c->pid,PROCESS_ERROR,c->identifier,s);
	   return(-1);
   }
   _dc->next = NULL;
   /* add operation to delayed queue */
   _dcc = delayedroot;
   if ( delayedroot != NULL )
   {
	   while(_dcc->next != NULL)
			_dcc = _dcc->next;
	   _dcc->next = _dc;
   }
   else delayedroot = _dc;
   _dc->oper = *c;
   _dc->request = t;
   return(ret);
}

/* void delete_delayed_operation()
 * Delete the delayed_operation queue
 * Application process are notified
 *
 * Argument : None
 *
 * Return : None
 */
void delete_delayed_operation(void)
{
struct delayed_operation  *_dc;
struct delayed_operation *_dcc;
int                       slot;

/* delete delayed_operation queue */
	_dc = delayedroot;
	while(_dc)
	{
		  if ( (slot = find_application_slot(_dc->oper.pid)) != NOT_AN_INDEX )
		  {
				ker_build_acknowledge(_dc->oper.pid,
						  PROCESS_ERROR,
						  _dc->oper.identifier,
						  "Wdx process terminated");
				com_close_fifo(applicationtable[slot].des);
			applicationtable[slot].use = SLOT_NOT_USED;
	  }
		  _dcc =  _dc->next;
		  free(_dc);
		  _dc = _dcc;
	}
}

/* int lock_opr()
 * Process a lock operation message : call the lock protocol
 *
 * Argument : pointer to operation
 *	      operation value
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int lock_opr(c,t)
struct operation *c;  /* operation */
enum lock_request t;  /* operation type */
{
int                   ret; /* returned value */
char		       *s; /* pt to message */
enum process_status    st; /* acknowledge status */



   ret = lck_lock_operation_input(c,t,&wdxsend);
   switch(ret)
   {
   case LOCK_D :
	s = "Command completed";
	st = PROCESS_SUCCESS;
		break;
   case LOCK_S :
	s = "Command processing";
	st = PROCESS_SUCCESS;
		break;
   case LOCK_T :
	s = "Too many resource managed";
	st = PROCESS_ERROR;
		break;
   case LOCK_M :
	s = "Not a local mount point";
	st = PROCESS_ERROR;
		break;
   case LOCK_P :
	s = "Not allowed for mount point";
	st = PROCESS_ERROR;
		break;
   case LOCK_R :
   default :
	s = "Command rejected by wdx lock protocol";
	st = PROCESS_ERROR;
	break;
   }
   ker_build_acknowledge(c->pid,st,c->identifier,s);
   return(0);
}

/* int control_opr()
 * Process a control operation message. Control value can be either
 * WDX_CONTROL_OPEN : open an application_slot if one available and connect
 * to the application
 * WDX_CONTROL_CLOSE : close the connection with the application then release the
 * application_slot
 *
 * Argument : pointer to operation
 *            operation type
 *
 * Return   : 0 ( status update ) ok
 *            1 ( no status update ) ok
 *            -1 fatal error
 */
int control_opr(c,t)
struct operation *c;
int               t;
{
int                    slot;  /* slot number */
int                       i;  /* index */
int                     des;  /* fifo descriptor */
struct application_slot *pt;  /* pointer to current application_slot */

   switch(t)
   {
   case WDX_OPEN_CONTROL : /* open communication link with wdx application */
	/* check if application slot available */
	slot = NOT_AN_INDEX;
	for(i=0;i<MAX_APPLICATION_ON_PROCESS;i++)
		if( applicationtable[i].use == SLOT_NOT_USED )
	{
		slot = i;
		break;
	}
	if ( (des = com_open_fifo(c->operationdata,WDX_PROCESS)) >= 0 )
	{
		 /* no more slot available : send error acknowledge + close fifo */
		 if ( slot == NOT_AN_INDEX )
		 {
				send_acknowledge(des,
				   PROCESS_ERROR,
				   c->identifier,
				   "No link free ");
		  PWARNING_B
		  "No more application link free "
		  PWARNING_E;
		  com_close_fifo(des);
		 }
		 /* available slot : store application parameters */
		 /* send control_first + acknowledge */
		 else
		 {
			  pt = &(applicationtable[slot]);
			  pt->use = SLOT_USED;
			  pt->des = des;
			  pt->pid = c->pid;
			  pt->mode = c->operationargument;
		  ker_build_control(slot,PROCESS_FIRST);
		  ker_build_acknowledge(c->pid,
						PROCESS_SUCCESS,
						c->identifier,
					"Connection done");

				  if ((locktable[slot].lock = spc_lookat_app_lock(c->operationdata,
										  ker_clean_up,
										  c->pid)) == -1)
				  {
			 PWARNING_B
			 "Can't look at the application lock"
				 PWARNING_E;
				  }
		  locktable[slot].pid = pt->pid;
		 }
	}
		else
	{
		 PWARNING_B
		 "FIFO open failed"
		 PWARNING_E;
		}
		break;
   case WDX_CLOSE_CONTROL : /* close communication link with wdx application */
	/* search for application_descriptor */
	slot = find_application_slot(c->pid);
	if ( slot != NOT_AN_INDEX )
	{
		ker_build_control(slot,PROCESS_LAST);
		ker_build_acknowledge(c->pid,
				  PROCESS_SUCCESS,
				  c->identifier,
				  "Connection closed");
		com_close_fifo(applicationtable[slot].des);
		reset_application_slot(slot);
	}
		break;
   default :
		sprintf(messagestring,
				"control_opr(), invalid control operation %d",
				t);
		PWARNING_B
		messagestring
		PWARNING_E;
		break;
   }
   return(0);
}

/* int command_message()
 * Process an input command message.
 * According to the command type one routine is selected.
 * For TREQUEST : reqevt event table is used
 * For TANSWER : ansevt event table is used
 * For TACTION : actevt event table is used
 *
 * Argument : pointer to command
 *
 * Return   : 0 ok
 *           -1 fatal error
 */
int command_message(c)
struct command *c;
{
struct event_table *tab; /* pt to right event table */
struct event_table *pt;  /* pt to event table */
int                ret;  /* returned value */


   if(c)
   {
	   switch(c->type)
	   {
	   case TREQUEST :
		tab = reqevt;    /* point to reqevt table */
		break;
	   case TANSWER :
		tab = ansevt;    /* point to answer table */
		break;
	   case TACTION :
		tab = actevt;    /* point to actevt table */
		break;
	   default :
			sprintf(messagestring,
			"command_message(), invalid command type %d",
			c->type);
			PWARNING_B
			messagestring
		PWARNING_E;
		return(0);
		break;
	   }
/* TRACE USED for PAUSE debugging
	sprintf(messagestring,"type : %d, value : %d",c->type,c->value1);
	PINFOR_B
	messagestring
	PINFOR_E;
*/
	   for(pt=tab;pt->value;pt++)
	   {
			if ( pt->value == c->value1 )
		{
				 ret = (pt->evt[substatus].funct)(c,
						pt->parameter1);
				 if ( ret == 0 )
				 {
				  status = pt->evt[substatus].status;
				  substatus = pt->evt[substatus].substatus;
			  activity_protocol(substatus);
		 }
				 break;
			}
	   }
   }
   if ( substatus == SUBFAULT || status == W_FAULT )
	return(1);
   else return(0);
}

/* int process_message()
 * Process an input message. On message type value select the right process
 * routines :
 *	- HEARTBEAT : wat_input()
 *		      lck_resource_input()
 *	- OPERATION : opertab is used to select the good routine
 *	- COMMAND   : command_message() switch is called.
 * First checksum is checked then the message is filtered on validity of the
 * header.
 * Before calling, lck_resource_input() and command_message() messages are
 * filtered according to datanumber value -> messages coming from two different
 * networks are processed only once.
 *
 * Argument : pointer to encoded_message em
 *
 * Return   : 0 ok
 *           -1 fatal error
 */
int process_message(inpt)
struct socket_input  *inpt;
{
struct encoded_message *em;
struct message m;
struct command *c;
struct operation *o;
struct heartbeat *h;
int index;
int datanumber;
int ret;
struct event_table *pt; /* pointer to oprevt table */


 if ( inpt )
 {
	em = (struct encoded_message *) inpt->buffer;
	memset(&m,'\0',sizeof(m));
	if(is_checksum_valid(em))
	{
		lib_decode_message(em,&m);
		if((ret = is_message_valid(&m)) > 0 )
		{
			index = m.head.memberindex;
			datanumber = m.head.datanumber;
			switch(m.head.datatype)
			{
			case HEARTBEAT :
				 h = &(m.m.h);
				 if(index != localindex)    /* remote message */
				 {
					  analyze++;
					  if (wat_input(&(h->wat),inpt->number) == -1)
						  return(-1);
					  if (validate_data_number(index,datanumber))
			  {
						  return(lck_resource_input(index,
						h->res,
						&wdxsend));
					  }
				 }
				 break;
			case COMMAND   :
		 c = &(m.m.c);
		 if(validate_data_number(index,datanumber))
			return(command_message(c));
		 break;
		case OPERATION :
		 o = &(m.m.o);
				 if(index == localindex)    /* local message only */
				 {
			for(pt=opertab;pt->value;pt++)
			{
			if ( pt->value == o->type && index == localindex )
			{
				ret = (pt->evt[substatus].funct)(o,
									pt->parameter1);

				if ( ret == 0 )
				{
				status = pt->evt[substatus].status;
				substatus = pt->evt[substatus].substatus;
				activity_protocol(substatus);
				}
				return(0);
			}
			}
			sprintf(messagestring,
						 "process_message(), invalid operation type %d",
						  o->type);
			PWARNING_B
			messagestring
			PWARNING_E;
				 }
				 break;

			default        :
				 sprintf(messagestring,
						 "process_message(), invalid message type %d",
						 m.head.datatype);
				 PWARNING_B
				 messagestring
				 PWARNING_E;
				 break;
			}
		}
		else return(ret);
	}
 }
 return(0);
}


/** CDM002 **/

/* void broadcast_test()
 * This routine is used to detect broadcast sending problems.
 *
 * if absolute value of ( current - previous ) is greater than
 * localbroadcast value, a message is logged
 *
 * Parameter : pt to previous timer value
 *             pt to current timer value
 *
 * Return    : nothing
 */

void broadcast_test(time_t *p, time_t *c)
{
 if  (( (long)*c - (long)*p ) > localbroadcast)
 {
	  PWARNING_B
	  "Broadcast sending problem detected"
	  PWARNING_E;
	  sprintf(messagestring,"Previous sending was %s",
			  lib_generation_to_s(p));
	  PCOMPLEMENT_B
	  messagestring
	  PCOMPLEMENT_E;
 }
}
/** CDM002 **/
	 
