/*
 *   WDX: WitchDoctor-WDX, versatile distributed lock manager
 *   Copyright (C) 2003 Dominique Chabord,
 *   Please refer to credits section below
 *
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   ************************************************
 *
 *   History and credits:
 *   First version of WitchDoctor-WDX was based on:
 *
 *   DECwatchdog/Autopilot, watchdog and lock manager
 *   Copyright (C) 2003 HP
 *
 *   Developed by
 *     Computer Special Systems, Annecy France
 *     A Division of Digital Equipment Corporation
 *   and
 *     Custom-Systems, Annecy France
 *     An engineering center of Compaq
 *
 *   Provided under GNU-GPL terms by HP, 30 June 2003
 *   GNU-GPL replaces previous license statements
 *
 *   Rebranded, ported and modified for Shaman-X project, 15 July 2003
 *   by Dominique Chabord
 *   and maintained by other contributors as mentionned in the code
 *
 *   ************************************************
 *
 *   WDX is promoted by Shaman-X project
 *
 *   Check for new versions and documentation at
 *   http://www.shaman-x.org
 *
 *   ************************************************
 *
 *   Most of WDX code is designed to run unchanged on
 *   different operating systems, including at V1.0 time:
 *   Gnu-Linux, WindowsNT, DigitalUNIX and OpenVMS.
 *   Operating system specifics are identified by dedicated files.
 *
 *   Evolutions
 *   ----------
 *   July-2003     Dominique Chabord    WitchDoctor-WDX V1.0
 *                 Creation
 *
 *   ************************************************
 *
 *
 *	wdx_lockprotocol.c
 *
 *	Module to process the lock protocol
 *
 *	Routines for administration of resource lock requests,
 *	sent by local wdx application to the local wdx process.
 *
 *      Public Routines :
 *              void lck_init()
 *		int  lck_ins_information_in_newlist()
 *		void lck_delete()
 *              int  lck_validate()
 *              void lck_purge_requester()
 *		void lck_reset_resource()
 *		void lck_activity_set()
 *              void lck_unmark_resource()
 *              int  lck_resource_input()
 *              int  lck_lock_operation_input()
 *              int  lck_wait_operation_input()
 *              int  lck_automatic_request_wish()
 *
 *	Local Routines :
 *
 *              void print_wishlist()
 *              int find_index_for_resource()
 *              enum lock_protocol  lock_protocol()
 *              enum lock_transition build_transition()
 *              enum lock_answer build_answer()
 *              enum lock_answer build_answer_to_local_request()
 *              enum resource_status build_status()
 *              int  insert_resource_in_newlist()
 *              int  add_request_to_rat()
 *              int  add_request_to_wish_list()
 *              int  add_answer_to_wish_list()
 *              int  add_transition_to_wish_list()
 *              void remove_request_from_rat()
 *              struct rat *purge_wish_list()
 *              void resource_transition()
 *		int add_to_wait_list()
 *              int update_resource_count()
 *              int update_resourcelist()
 *              int request_input()
 *              int answer_input()
 *              int transition_input()
 *              int information_input()
 *              void set_not_cleaned()
 *              void set_cleaned()
 *              void reset_not_cleaned()
 *		int validate_wait_list()
 *		int validate_wish_list()
 *              void protocol_trace()
 *              void protocol_trace_denied()
 *		void protocol_trace_conflict()
 *		void lock_notification()
 *		int add_identifier_to_rat()
 *		int operation_preprocess()
 *
 ************************************************************************/
/*
 *
 *  27-Apr-94 -- Cornu
 *      Creation
 * MK001 08-JUN-95 M.Kennealy Modifications for i386 compilation
 */

/* Abstract :
 * The lockprotocol module is in charge of the administration of the 
 * resource lock requests.
 *
 * Each resource request is inserted in a list called the wishlist. This list
 * includes for each resource name one subqueue called the rat :
 *
 *   res1 -> res2 -> res3 -> NULL
 *    |       |       |
 *    v       v       v
 *   rat11   rat21   rat31
 *    |       |       |
 *    v       v       v
 *   rat12   NULL    rat32
 *    |               |
 *    v               v
 *   NULL            rat33
 *                    | 
 *                    v
 *                   NULL
 *
 * Each 'rat' element represent a request for a resource. We consider two types 
 * of request :
 *              - local request : the request initiator is an application of the
 *                local process ( or the local process itself )
 *              - remote request : the request initiator is a remote
 *                application or process.
 *
 * For 'local' rat request ( lock_operation_input() routine ), we must do the 
 * following :
 *              - broadcast the request
 *              - wait for the answers ( answer_input() routine )
 *              - process the lock protocol when all the answers are received
 *                or when the timeout has reached LOCK_TIMEOUT.
 *                ( validate_wish_list() routine )
 *              - send a transition message in all cases :
 *                      - with transition value if the resource status changes.
 *                      - with NOTRANS if the resource status does not change
 *              - purge the rat request when a decision is taken
 *
 * For 'remote' rat request ( request_input() routine ), we must do the 
 * following :
 *              - broadcast an answer according to the lock protocol table
 *              - wait until a transition is received ( transition_input()
 *                routine() ) or the timeout has
 *                reached TRANSITION_TIMEOUT ( validate_wish_list() routine )
 *              - purge the rat request when a transition is received or when
 *                the timeout has reached TRANSITION_TIMEOUT.
 *
 * So we can receive three types of messages :
 *              - request message : 'rat' creation
 *              - answer message : 'rat' update
 *              - transition message : 'rat' update
 *
 * The lock protocol manage the following collisions :
 *      1/ A remote request input and a local request input : the request
 *         with lowest initiator value is processed. The other is rejected.
 *      2/ A local request input while a remote request is waiting for a
 *         transition : the local request is delayed until the remote request
 *         has been finished.
 *      3/ A local request input and a local request input : the second
 *         local request input is rejected.
 *
 *
 * Local request input are transmitted from processes called "application". 
 * Applications are notified of request execution by notification events.
 * These events are built in the wdx_kernel module.
 *
 * BE CAREFULL : The routine lck_init should called first to set
 * static local pointers. This routine can be called using the structure
 * block 'function_table'.
 */

#undef  WDX_LOCKPROTOCOL_MODULE
#define WDX_LOCKPROTOCOL_MODULE
#include	"wdx.h"

#define AUTOMATIC_IDENTIFIER -1	/* used for request passed through automatic 
				   wish */
extern int automaticflag;	/* automatic flag request */

/* lock_status is used to application notification */
enum lock_status { LOCK_ERROR, LOCK_SUCCESS };

enum deny_event { DENY_MESSAGE, DENY_TIMEOUT };

/* possible values for delay */
enum delay { NOTDELAYED, DELAYED };

/* possible values for sent */
enum request_sent { NOTSENT, SENT };

/*
 * The wishlist structure is used to perform the locking protocol
 */

struct wish_list
{
    struct wish_list    *previous; /* pointer to previous list element */
    struct wish_list        *next; /* pointer to next list element */
    char        name[RESOURCE_LEN]; /* resource name */
    struct rat             *queue; /* poiter to r.a.t. queue */
};

/* identifier structure : defines the identifier of an application request */

struct identifier
{
    pid_t               pid;	/* application pid */
    int		 identifier;    /* request identifier */
    struct identifier *next;    /* pt to next queue element */
};

struct rat
{
    struct rat              *previous; /* previous element of queue */
    struct rat                  *next; /* next element of queue */
    int                     initiator; /* index of the initiator of the req. */
    struct identifier      *identlist; /* identifier list for the request */
    struct wdx_timer            timer; /* timer for the request */
    enum delay                   flag; /* NOTDELAYED or DELAYED */
    enum request_sent		 sent; /* SENT or NOTSENT */
    enum lock_request              lr; /* lock request sent by the initiator */
    enum lock_answer la[MEMBER_COUNT]; /* lock answer table (for local req.) */
    enum lock_transition           lt; /* transition for local request */  
    time_t                    generation; /* generation of the lock */
};

/* BE CAREFULL : each structure of the following must fill into the 3
 * fields called value1, value2, value3 of the resource structure
 */
struct information
{
    enum resource_status  status;    /* status of the resource */
    int                initiator;    /* request initiator of the resource */
    int               generation;    /* generation of the lock */
};

struct shortcut
{
    enum resource_status  status;    /* status of the resource */
    int                initiator;    /* request initiator of the resource */
    time_t               generation;    /* generation of the lock */
};

struct request
{
    enum lock_request     lr;    /* lock request value */
    int                   ri;    /* requester index */
    int                   gn;    /* last generation number */
};

struct answer
{
    enum lock_answer      la; /* lock answer value */
    int                   ri; /* requester index to select the right request */
    int                   ai; /* answerer index to update the right answer */
};

struct transition
{
    enum lock_transition lt;  /* lock transition value */
    int                  ri;  /* requester index to select the right request */
    int                  gn;  /* generation of the lock */
};

#define LEN 40

struct api_status
{
	enum process_status success;
	enum process_status error;
	enum resource_status waited;
        char success_message[LEN];
        char error_message[LEN];
};				

struct api_status api_statuss[] =
{
        {  0, 0, NOTKNOWN, "", "" },
        { PROCESS_SUCCESS, PROCESS_ERROR, FREE, "%s declared", 
	  "Can not declare %s " },
        { PROCESS_SUCCESS, PROCESS_ERROR, NOTKNOWN, "%s erased" , 
	  "Can not erase %s" },
        { PROCESS_SUCCESS, PROCESS_ERROR, RESERVED, "%s reserved", 
	  "Can not reserve %s" },
        { PROCESS_SUCCESS, PROCESS_ERROR, MOUNTED, "%s mounted", 
	  "Can not mount %s" },       
        { PROCESS_SUCCESS, PROCESS_ERROR, FREE, "%s freed", 
          "Can not free %s" },        
        { PROCESS_SUCCESS, PROCESS_ERROR, FAULT, "%s faulted", 
          "Can not fault %s" },
        { PROCESS_SUCCESS, PROCESS_ERROR, RESERVED, "%s reserved", 
          "Can not get %s" }
};

/* The wait_list structure is used to store the pid of the waiting application
 * process. An element is added to the queue when a WAIT_FOR_RESERVE or
 * WAIT_FOR_MOUNT operation is entered.
 * An element is removed when the lock has been reached and a notification is
 * sent back.
 */
struct wait_list
{
    struct wait_list *next;
    struct wait_list *previous;
    pid_t  pid;			/* request pid */
    int    identifier;		/* request identifier */
    enum wait_operation    wait;	/* wait value */
    char name[RESOURCE_LEN+1];  /* resource name */
};

/* The mismatch_list structure is used to saved information concerning mismatch
 * between members. Mismatch appears when a double lock on a resource is
 * persistent due to network dissociation
 */

struct mismatch_list
{
    int              initiator; /* mismatch initiator */
    char  name[RESOURCE_LEN+1]; /* resourcename */
    enum resource_marked alive; /* used as mismatch alive flag */
    struct wdx_timer     timer; /* timer value */
    struct mismatch_list *next; /* pt to next element in list */
    struct mismatch_list *previous; /* pt to previous element in list */
};

/* lock protocol timeout */
#define LOCK_TIMEOUT              1
#define TRANSITION_TIMEOUT       -1

#define LOCAL 0
#define REMOTE 1

/* possible values taken by the lock protocol */
enum lock_protocol { NODECISION, WDXUNLOCK, WDXLOCK };

static struct wish_list *wishroot = NULL;
static struct wait_list *waitroot = NULL;
static struct mismatch_list *mismatchroot = NULL;

/* pointer to general data */
static struct member *mbr;              /* pt to memberlist */
static struct resource_def *rsd;        /* pt to resourcelist */
static struct group *grp;		/* pt to group base */

/* silence flag : if set to NO then no output for the lockprotocol */
static enum mask lockactivity;
/* script flag : if set to NO then no script triggered by the lockprotocol */
static enum mask scriptactivity;

/* lockprotocol timeout */
static int locktimeout;		    

/* wishlist count and resource count */
static int *membercountpt;		/* number of members in group */
static int wishlistcount;
static int *resourcecountpt;		/* number of resources */

/* broadcastflag : 0 no data in newlist / nonzero data in newlist */
static int *broadcastflag;              /* pt to broadcast flag */
static struct resource *new;            /* pt to message resource list */

/* Local routines prototypes */
void print_wish_list();
int find_index_for_resource(char *res);
enum lock_protocol lock_protocol(struct rat *r);
enum lock_transition build_transition(enum lock_request req);
enum lock_answer build_answer(char *res,enum lock_request event);
enum lock_answer build_answer_to_local_request(char *res,enum lock_request event);
enum resource_status build_status(enum lock_transition req);
int insert_resource_in_newlist(char *res,enum event_type event,char *data);
int add_request_to_rat(struct rat **prat,struct request *r,enum delay del,char *res);
int add_request_to_wish_list(char *res,struct request *r, struct rat **currenpt);
int add_answer_to_wish_list(char *res,struct answer *a,struct rat **currentpt);
int add_transtion_to_wish_list(char *res,struct transition*t,struct rat **currentpt);
void remove_request_from_rat(struct wish_list *w,struct rat *q);
struct rat *purge_wish_list(char *r,int param);
void resource_transition(char *rn,int tr, enum resource_status os, enum resource_status ns);
int add_to_wait_list(pid_t pid,int identifier, enum wait_operation wait, char *res);
int update_resource_count();
int update_resourcelist(char *res,enum resource_status sta,int own,int gen);
int request_input(char *res,struct request *lr);
int answer_input(char *res, struct answer *la);
int transition_input(char *res,struct transition *lt);
int information_input(char *res,struct information *li);
void set_not_cleaned();
void set_cleaned(int id,char *name);
void reset_not_cleaned(int id);
int validate_wish_list(int decrement,int *brdflg);
void validate_delayed_event(int dec);
int validate_wait_list();
void protocol_trace(enum event_type event,int event_value,char *res,int index);
void protocol_trace_denied(enum deny_event deny,enum lock_request r,char *res,int index);
void protocol_trace_conflict(int conflit,char *res);
void lock_notification(struct rat *r,char *n,enum lock_status s);
int add_identifier_to_rat(struct rat *c,pid_t p,int i);
int operation_preprocess(struct operation *c,enum lock_request t, int index,int *r);
void mismatch_mark(char *r,int i);
void mismatch_unmark();
void mismatch_reset(int i);
void mismatch_validate(int d);
int shortcut_input(char *res, struct shortcut *lo);  /* MK001 */


/*
 * PUBLIC ROUTINES : prefixe is "lck"
 */

/* void lck_init()
 * Initialize mbr and rsd pointer used by lockprotocol module
 * For COLD_INIT or COLD_REINIT then resource table are resetted
 * 
 * Argument : COLD_INIT or COLD_REINIT or WARM_REINIT
 *	      pointer to group structure
 *            pointer to built_in resource message table
 *
 * Return : None
 */
void lck_init(s,g,m)
enum startup_mode s;   /* mode */
struct group     *g;   /* group pointer */
struct message   *m;   /* message pointer */
{
int                    _i; /* index */
struct resource *_newlist; /* pointer to resource table used to send mess. */

     /* generic pointer init */
     _newlist      = m->m.h.res;
     mbr           = g->memberlist;
     rsd           = g->resourcelist;
     new           = _newlist;
     grp           = g;
     membercountpt   = &(g->membercount);
     resourcecountpt = &(g->resourcecount);

     /* reset activity flags */
     lockactivity   = NO;
     scriptactivity = NO;

     /* locktimeout init : max of broadcast time + 1 */
     locktimeout = 0;
     for(_i=0;_i<g->membercount;_i++)
	if((g->memberlist[_i].broadcast)>locktimeout)
	    locktimeout = g->memberlist[_i].broadcast;
     locktimeout++;

     /* on startup mode init */
     switch(s)
     {
     case COLD_INIT :	/* cold init */
     case COLD_REINIT :	/* cold reinit */
        /* mastertoken init */
	strncpy(mastertoken,MASTER_TOKEN,RESOURCE_LEN);
	strncpy((rsd+MASTER_INDEX)->name,MASTER_TOKEN,RESOURCE_LEN);
	(rsd+MASTER_INDEX)->status = FREE;
	(rsd+MASTER_INDEX)->initiator = NOT_AN_INDEX;
	(rsd+MASTER_INDEX)->generation = 0;
        /* wishlistcount and resourcecount init */
	wishlistcount = 0;
	*resourcecountpt = 0;
	/* reset resource list */
	for(_i=FIRST_RESOURCE;_i<RESOURCE_COUNT;_i++)
	{
	    (rsd+_i)->name[0] = '\0';
	    (rsd+_i)->status = NOTKNOWN;
	    (rsd+_i)->initiator = NOT_AN_INDEX;
	    lib_timer_reset(&((rsd+_i)->delay));
	}
	break;
     case WARM_REINIT :   /* warm init : restart from snapshot */
	break;
     }
}

/* int lck_ins_information_in_newlist()
 * For all information message in resource status table with flag NOT_MARKED, 
 * insert information message in built_in resource table
 *
 * Argument : Pt to broadcast flag 
 * 
 * Return   : -1 on error
 *            0 if ok
 */
int lck_ins_information_in_newlist(flagpt)
int *flagpt;
{
int _x;
int _ret;
int _insert;
struct resource_def *_r;
struct information _i;
struct request _req;
struct rat *_current;

   broadcastflag = flagpt;
   for(_x=0;_x<RESOURCE_COUNT;_x++)
   {
         _r = rsd + _x;
         if ( ( _r->name[0] != '\0' ) && ( _r->flag == NOT_MARKED ) )
         {
              _insert = 0;
              switch(_r->status)
              {
              case FREE:
                  _insert++;
                  break;
              case FAULT:
                  if ( _r->initiator == localindex )
                       _insert++;
                  break;
              case MOUNTED:
                  if ( _r->initiator == localindex )
                  {
                       if(spc_check_mount_point(_r->name,WDX_PROCESS) != 0)
                       {
                           _req.lr = FREEING;
                           _req.ri = localindex;
                           _req.gn = _r->generation;
                           _ret = add_request_to_wish_list(_r->name,
                                                          &_req,
                                                          &_current);
                           if ( (!_ret) && (_current) )
                           {
                                if ( _current->flag == NOTDELAYED )
                                {
                                     _current->la[localindex] = 
                                           build_answer_to_local_request(
                                                _r->name,
                                                _req.lr);
                                     _ret = insert_resource_in_newlist(
                                                _r->name,
                                                REQUEST,
                                                (char *)&_req);
                                }
                            }
                       }
                       else _insert++;
                  }
                  break;
              case RESERVED:
                  if ( _r->initiator == localindex )
                       _insert++;
                  break;
              }
              if ( _insert )
              {
                   _i.generation  = _r->generation;
                   _i.status = _r->status;
                   _i.initiator = _r->initiator;
                   _ret = insert_resource_in_newlist(_r->name,
						     INFORMATION,
						     (char *)&_i);
                   if ( _ret == -1 )
                        return(-1);
              }
        }
   }          
   return(0);
}

/* void lck_delete()
 * delete all requests from the wish_list and wait_list and mismatch_list
 *
 * Argument : none
 *
 * Return   : none
 */
void lck_delete(void)
{
struct wish_list *_p;		
struct wish_list **_pp;	/* address of the preceding pointer */
struct rat *_q;
struct rat *_qn;
struct wait_list *_pw;
struct wait_list *_pnw;
struct identifier *_i;
struct identifier *_in;
struct mismatch_list *_m;
struct mismatch_list *_mw;

//syslog(LOG_ERR,"lck_delete has been called");
/* delete mismatch_list queue */
   _mw = mismatchroot;
   while(_mw)
   {
     _m = _mw;
     _mw = _mw->next;
     free(_m);
   }

/* delete wait_list queue */
   _pw = waitroot;
   while(_pw)
   {
       ker_build_notification(_pw->pid,
			      PROCESS_ERROR,
			      _pw->identifier,
		              0,
			      NOTKNOWN,
			      "Wdx process terminated");
       _pnw = _pw->next;
       free(_pw);
       _pw = _pnw;
   }

/* delete wish_list queue */
    _pp = &wishroot;
    _p  = *_pp;
    while (_p)	/* for all elements in wish_list */
    {
        _qn = _p->queue;
	/* free rat list */
        for ( _q = _p->queue; _q; _q = _qn)
        {
	      /* free identlist */
	      _in = _q->identlist;
	      for(_i = _q->identlist;_i;_i = _in)
	      {
	          if ( _i->pid != getpid() )
	          {
	               ker_build_notification(_i->pid,
	    		                      PROCESS_ERROR,
	  		                      _i->identifier,
			                      0,
			                      NOTKNOWN,
			                      "Wdx process terminated");
		  }
	          _in = _i->next;
	          free(_i);
	      }
              _qn = _q->next;
              free(_q);     
        }
        *_pp = _p->next; /* delete link */
        free(_p);
	_p  = *_pp;      /* advance pointer */
        /* decrement wishlistcount */
        wishlistcount--;
    }
        
}

/* int lck_validate()
 * Validate wait list and wish list.
 *
 * Argument : decrementation value
 *            pointer to broadcastflag
 */
int lck_validate(int decrement,int *brdflgpt)
{
   validate_wait_list();
   validate_delayed_event(decrement);
   mismatch_validate(decrement);
   return(validate_wish_list(decrement,brdflgpt));
}

/* void lck_purge_requester()
 * For an requester index, change all locked resource to FREE.
 * Do not purge wish_list
 *
 * Argument : requester index
 *	      member status
 *	      timeout value
 *
 * Result   : None
 */
void lck_purge_requester(ow,st,tm)
int ow;
enum member_status st;
int        tm;
{
int _i;
struct eventscript_data evt; /* data used to call lib_eventscript */


   if(ow > NOT_AN_INDEX && ow < *membercountpt )
   {
        for(_i=0;_i<RESOURCE_COUNT;_i++)
           if ( (rsd+_i)->initiator == ow )
           {
                if ((rsd+_i)->status == RESERVED || (rsd+_i)->status == MOUNTED ||
                     (rsd+_i)->status == FAULT )
                {
		    if ( st == MASTERHOLDER )
		    {
			(rsd+_i)->oldstatus = (rsd+_i)->status;
			(rsd+_i)->oldinit = ow;    /* search if the mismatch is already registered */
			lib_timer_set(&((rsd+_i)->delay),tm);
	                evt.script = NO;
		    }
		    else
		    {
			evt.script = scriptactivity;
		    }
	            evt.type = EVT_RESOURCE; 
	            evt.display = YES;
	            evt.index = ow;
	            evt.group = grp;
	            strcpy(evt.data.r.name,(rsd+_i)->name);
	            evt.data.r.old = (rsd+_i)->status;
	            evt.data.r.new = FREE;
	            lib_eventscript(&evt);     
                    (rsd+_i)->status = FREE;
                    (rsd+_i)->initiator = ow;
                }
           }
   }
}

/* void lck_purge_resource_delayed()
 * A new masterholder appears, trigger script for each delayed eventscript.
 *
 * Argument : None
 *
 * Result   : None
 */
void lck_purge_resource_delayed()
{
int _i;
struct eventscript_data evt; /* data used to call lib_eventscript */


   for(_i=0;_i<RESOURCE_COUNT;_i++)
       if ( lib_timer_alive(&(rsd+_i)->delay) )
   {
		lib_timer_reset(&((rsd+_i)->delay));
		evt.script = scriptactivity;
	        evt.type = EVT_RESOURCE; 
	        evt.display = NO;
	        evt.index = (rsd+_i)->oldinit;
	        evt.group = grp;
	        strcpy(evt.data.r.name,(rsd+_i)->name);
	        evt.data.r.old = (rsd+_i)->oldstatus;
	        evt.data.r.new = FREE;
	        lib_eventscript(&evt);     
   }
}

/* void lck_reset_resource()
 * For a resource wich does not have any owner, transition to FREE state
 * if locked.
 * Argument : None
 *
 * Result   : None
 */
void lck_reset_resource()
{
int _i;

   for(_i=0;_i<RESOURCE_COUNT;_i++)
   {
       if ( (rsd+_i)->initiator == NOT_AN_INDEX )
       {
            if ((rsd+_i)->status == RESERVED || (rsd+_i)->status == MOUNTED ||
                (rsd+_i)->status == FAULT )
            {
                resource_transition((rsd+_i)->name,NOT_AN_INDEX,
                                    (rsd+_i)->status,FREE);
                (rsd+_i)->status = FREE;
                (rsd+_i)->initiator = NOT_AN_INDEX;
           }
       }
   }
}

/* void lck_activity_set()
 * Set lockactivity and lockscript to the values passed as argument
 *
 */
void lck_activity_set(lockvalue,scriptvalue)
enum mask  lockvalue; /* to set or reset lockactivity */
enum mask scriptvalue; /* to set or reset scriptactivity */
{
   /* set lockactivity activity */
   lockactivity   = lockvalue;
   scriptactivity = scriptvalue;
}

/* void lck_unmark_resource()
 * Set flag of all resources to NOT_MARKED
 *
 * Argument : None
 *
 * Return : None
 */
void lck_unmark_resource()
{
int _x;

   for ( _x=0;_x<RESOURCE_COUNT;_x++)
         (rsd+_x)->flag = NOT_MARKED;
}

/* int lck_resource_input()
 * Process all the "res" table of the heartbeat messages
 * The "res" table is a table of RESOURCE_COUNT elements of type "resource"
 *
 * Argument    : index of the initiator
 *               pointer to resource table
 *               pointer to broadcastflag
 *
 * Return      : 0 ok
 *               -1 fatal error
 */
int lck_resource_input(id,rt,brdflgpt)
int               id; /* id of the member which has sent this table */
struct resource rt[]; /* pointer to res table */
int        *brdflgpt; /* pointer to broadcast flag */
{
int              _x;     /* index */
struct resource *_r;     /* resource element */
struct request *_lr;     /* used to read request input */
struct answer  *_la;     /* used to read answer input */
struct transition *_lt;  /* used to read transition input */
struct information *_li; /* used to read information input */
struct shortcut *_lo;    /* used to read shortcut input */
int _ret;                /* returned value */

   /* set static value for broadcastflag to brdflgpt */
   broadcastflag = brdflgpt;
   
   /* initialize all resources to NOT_CLEANED */
   set_not_cleaned();
   /* initialize all mismatch elements to NOT_MARKED */
   mismatch_unmark();

   _ret = 0;

   if(rt)
   {
       for(_x=0;_x < RESOURCE_COUNT;_x++)
       {
           _r = rt + _x;
           if ( _r->name[0] != '\0' )
           {
                set_cleaned(id,_r->name);
                switch(_r->evttype)
                {
                case REQUEST :
                     _lr = (struct request *) &(_r->value1);      
                     _ret = _ret | request_input(_r->name,_lr);
                     break;
                case ANSWER :
                     _la = (struct answer *) &(_r->value1);      
                     _ret = _ret | answer_input(_r->name,_la);
                     break;
                case TRANSITION :
                     _lt = (struct transition *) &(_r->value1);      
                     _ret = _ret | transition_input(_r->name,_lt);
                     break;
                case INFORMATION :
                     _li = (struct information *) &(_r->value1);      
                     _ret = _ret | information_input(_r->name,_li);
                     break;
                case SHORTCUT :
                     _lo = (struct shortcut *) &(_r->value1);
                     _ret = _ret | shortcut_input(_r->name,_lo);
                     break;
                default :
                     sprintf(messagestring,
                             "resource_input(), invalid evttype %d",
                             _r->evttype);
                     PWARNING_B
                     messagestring
                     PWARNING_E;
                     break;
                }
           }
       }
   }
   /* cleaning up resources */
   reset_not_cleaned(id);
   /* cleaning up mismatch list */
   mismatch_reset(id);
   return(_ret);
}            

/* int lck_lock_operation_input()
 * Process a lock operation message 
 * 
 * Argument    : pointer to operation element
 *		 request value
 *               address of broadcast flag
 *
 * Return      : 0 ok
 *               -1 fatal error
 */
int lck_lock_operation_input(c,t,brdflgpt)
struct operation   *c;     /* operation pointer */
enum lock_request t;     /* request value */
int       *brdflgpt;     /* broadcast flag pointer */
{
struct request _r;
struct rat *_current;
int _ret;
int _i;

    _ret = 0;

    /* set static value for broadcastflag to brdflgpt */
    broadcastflag = brdflgpt;
    
    if(c)
    {
       _i = lib_index_of_resource(rsd,c->operationdata);
       if ( operation_preprocess(c,t,_i,&_ret) == -1 )
	    return(_ret);
       else
       {
            _r.lr = t; 
            _r.ri = c->index;
            if (_i = NOT_AN_INDEX )
                 _r.gn = 0;                     /* new resource */
            else _r.gn = (rsd +_i)->generation; /* existing resource */
            _ret = add_request_to_wish_list(c->operationdata,
                                           &_r,&_current);
	    if ( (_ret) || (!_current) )
		return(LOCK_R);
            _ret = add_identifier_to_rat(_current,c->pid,c->identifier);
	    if ( _ret )
		return(LOCK_R);
            if ( ( _current->initiator == localindex ) && 
                 ( _current->flag == NOTDELAYED ) && 
		 ( _current->sent == NOTSENT ) )
            {
                 _current->la[localindex] = 
                       build_answer_to_local_request(c->operationdata,_r.lr);
		 _current->sent = SENT;
		 protocol_trace(ANSWER,(int)_current->la[localindex],
                                      c->operationdata,_current->initiator);
                 _ret = insert_resource_in_newlist(c->operationdata,
                                                   REQUEST,
                                                  (char *)&_r);
            }
	    return(LOCK_S); /* operation processing */
       }
    }
    return(_ret);
}

/* int lck_wait_operation_input()
 * Process a wait operation message
 * 
 * Argument    : pointer to operation element
 *		 wait operation
 *
 * Return      : 0 ok
 *               -1 fatal error
 */
int lck_wait_operation_input(c,t)
struct operation *c;       /* operation pointer */
enum wait_operation t;     /* wait operation */
{

    if(c)
    {
	return(add_to_wait_list(c->pid,
	  		        c->identifier,
			        t,
			        c->operationdata) );
    }
    else return(-1);
}

/* int lck_automatic_request_wish()
 * Build a local RESERVING request for a resource
 * The resource has already been DECLARED.
 *
 * Argument : pt to resource name
 *            request
 *
 * Return   : 0 ok / -1 error
 */
int lck_automatic_request_wish(res,brdflgpt)
char *res;
int *brdflgpt;
{
struct request _r;
struct rat *_current;
int _ret;
int _i;
int _index;

    /* rejected if lockactivity set */
    if ( lockactivity == NO )
	return(-1);

    /* set static value for broadcastflag to brdflgpt */
    broadcastflag = brdflgpt;
    
    if(res)
    {
       _r.lr = RESERVING;
       _r.ri = localindex;
       if ( (_i = lib_index_of_resource(rsd,res)) == NOT_AN_INDEX )
            return(-1);
       _index = NOT_AN_INDEX;
       if ( strcmp(res,MASTER_TOKEN) == 0 )
       {
            _index = MASTER_INDEX;
            if ( (new + _index)->name[0] != '\0' )
                 return(-1);
       }
       else 
       {        
            for(_index=FIRST_RESOURCE;_index<RESOURCE_COUNT;_index++)
            {
                if ( (new +_index)->name[0] == '\0' )
                      break;
                /* is resource already listed in message ? */
                else if(strcmp(res,(new + _index)->name) == 0 )
                     return(-1);
            }
        }
           
       _r.gn = (rsd +_i)->generation;
       _ret = add_request_to_wish_list(res,&_r,&_current);
       if ( (!_ret) && (_current) )
       {
            _ret = add_identifier_to_rat(_current,getpid(),AUTOMATIC_IDENTIFIER);
            if ( ( _current->initiator == localindex ) && 
                 ( _current->flag == NOTDELAYED ) &&
		 ( _current->sent == NOTSENT ) )
            {
                 _current->la[localindex] = 
			build_answer_to_local_request(res,_r.lr);
		 _current->sent = SENT;
		 protocol_trace(ANSWER,(int)_current->la[localindex],res,
			        _current->initiator);
                 _ret = insert_resource_in_newlist(res,
                                                  REQUEST,
                                                  (char *)&_r);
            }
       }
    }
    return(_ret);
}

/*
 * LOCAL ROUTINES : no prefixe
 */

/* int find_index_for_resource()
 * Look into the resource table to find either :
 *      - the current index of a resource
 *      - the first free index.
 *
 * Argument : pt to res name
 *  
 * return   : index or NOT_AN_INDEX
 */
int find_index_for_resource(res)
char *res;
{
int _index;

   _index = NOT_AN_INDEX;
   if ( strcmp(res,MASTER_TOKEN) == 0 )
        _index = MASTER_INDEX;
   else 
   {        
        _index = lib_index_of_resource(rsd,res);
        if ( _index == NOT_AN_INDEX )
        {      
            for(_index=FIRST_RESOURCE;_index<RESOURCE_COUNT;_index++)
            {
                 if ( (rsd + _index)->name[0] == '\0' )
                      break;
            }
        }
   }           
   if ( _index == NOT_AN_INDEX )
   {
        PERROR_B
        "Find_index_for_resource(), too much resources managed by lock protocol"
        PERROR_E;
   }
   return(_index);
}

/* enum lock_protocol lock_protocol(r)
 * Performs the lock protocol on a rat element
 * Look to the answer table to determine the status of a request
 *
 * Argument : pointer to rat element
 *
 * Return   : WDXLOCK if the resource can locally be locked
 *            WDXUNLOCK if the lock is rejected
 *            NODECISION if the lockprotocol has not enough information
 */
enum lock_protocol lock_protocol(r)
struct rat *r;
{
int _i;
int _index;

   _index = 0;
   for( _i=0;_i < *membercountpt; _i++)
        {
             switch(r->la[_i])
                  {
                  case NOANSWER :
                       if ( lib_timer_reached(&(r->timer),LOCK_TIMEOUT) )
                            _index++;
                       else {
                            if ( !lib_is_member_valid(mbr,_i) )
                                 _index++;
                            else return(NODECISION);
                            }
                       break;
                  case OK :
                       _index++;
                       break;             
                  case NONOK :
                       return(WDXUNLOCK);
                       break;
                  default :
                       return(WDXUNLOCK);
                       break;
                  }
        }
    if ( _index == *membercountpt )
         return(WDXLOCK);
}

/* enum lock_transition build_transition()
 * Converts a request value into a transition value
 * 
 * Argument : lock_request
 * 
 * Return   : lock_transition
 */
enum lock_transition build_transition(req)
enum lock_request req;
{
    switch(req)
       {
       case DECLARING :
            return(DECLARE);
            break;
       case ERASING :
            return(ERASE);
            break;
       case RESERVING :
            return(RESERVE);
            break;
       case MOUNTING :
            return(MOUNT);
            break;
       case FREEING :
            return(FREET);
            break;
       case FAULTING :
            return(FAULTT);
            break;
       case GETTING :
            return(GETT);
            break;
       default :
            return(NOTRANS);
            break;
       }
}

/* enum lock_answer build_answer()
 *
 * Argument : Pointer to resource name
 *            lock_request
 * 
 * Return   : lock_answer
 */
enum lock_answer build_answer(res,event)
char *res;      /* resource name */
enum lock_request event;
{
int _index;

    _index = lib_index_of_resource(rsd,res);
    
    switch(event)
       {
       case DECLARING :
            if ( _index == NOT_AN_INDEX )
                 return(OK);
            else if ( (rsd + _index)->status == FREE )
                 return(OK);
            else return(NONOK);
            break;
       case ERASING :
            if ( _index == NOT_AN_INDEX )
                 return(OK);
            else if ( (rsd + _index)->status == FREE )
                 return(OK);
            else return(NONOK);
            break;
       case RESERVING :
            if ( _index == NOT_AN_INDEX )
                 return(NONOK);
            else if ( (rsd + _index)->status == FREE )
                 return(OK);
            else return(NONOK);
            break;
       case GETTING :
            if ( _index == NOT_AN_INDEX )
                 return(NONOK);
            else if ( (rsd + _index)->status == MOUNTED )
                 return(NONOK);
            else return(OK);
            break;
       case MOUNTING :
            if ( _index == NOT_AN_INDEX )
                 return(NONOK);
            else if ( (rsd + _index)->status == FREE )
                 return(OK);
            else return(NONOK);
            break;
       case FREEING :
            if ( _index == NOT_AN_INDEX )
                 return(NONOK);
            else return(OK);
            break;
       case FAULTING :
            return(OK);
            break;
       }
}

/* enum lock_answer build_answer_to_local_request()
 * Check if the current request is valid, according to local informations.
 *
 * Argument : Pointer to resource name
 *            lock_request
 * 
 * Return   : lock_answer
 */
enum lock_answer build_answer_to_local_request(res,event)
char *res;      /* resource name */
enum lock_request event;
{
int _index;

    _index = lib_index_of_resource(rsd,res);
    
    switch(event)
       {
       case DECLARING :
            if ( _index == NOT_AN_INDEX )
                 return(OK);
            else if ( (rsd + _index)->status == FREE )
                 return(OK);
            else return(NONOK);
            break;
       case ERASING :
            if ( _index == NOT_AN_INDEX )
                 return(OK);
            else if ( (rsd + _index)->status == FREE )
                 return(OK);
            else return(NONOK);
            break;
       case RESERVING :
            if ( _index == NOT_AN_INDEX )
                 return(NONOK);
            else if ( (rsd + _index)->status == FREE )
                 return(OK);
            else return(NONOK);
            break;
       case GETTING :
            if ( _index == NOT_AN_INDEX )
                 return(NONOK);
            else if ( (rsd + _index)->status == MOUNTED )
                 return(NONOK);
            else return(OK);
            break;
       case MOUNTING :
            if ( _index == NOT_AN_INDEX )
                 return(NONOK);
            else if ( (rsd + _index)->status == FREE &&
                      spc_check_mount_point(res,WDX_PROCESS) == 0 )
                 return(OK);
            else return(NONOK);
            break;
       case FREEING :
            if ( _index == NOT_AN_INDEX )
                 return(NONOK);
            else if ( (rsd + _index)->initiator == localindex )
                 return(OK);
            else return(NONOK);
            break;
       case FAULTING :
            return(OK);
            break;
       }
}

/* enum resource_status build_status()
 * Converts a transition value into a status value
 * 
 * Argument : lock_transition
 * 
 * Return   : resource_status
 */
enum resource_status build_status(req)
enum lock_transition req;
{
    switch(req)
       {
       case DECLARE :
            return(FREE);
            break;
       case ERASE :
            return(NOTKNOWN);
            break;
       case RESERVE :
            return(RESERVED);
            break;
       case MOUNT :
            return(MOUNTED);
            break;
       case FREET :
            return(FREE);
            break;
       case FAULTT :
            return(FAULT);
            break;
       case GETT :
            return(RESERVED);
            break;
       default :
            return(NOTKNOWN);
            break;
       }
}

/* int insert_resource_in_newlist()
 * Insert a request, answer , information or transition message in the 
 * new table used in message sent.
 * Before inserting a message check, mark the resource list element.
 *
 * Argument : Resource name
 *            Eventtype
 *            Pointer to data segment
 * 
 * Return   : 0 ok
 *            -1 fatal error
 */
int insert_resource_in_newlist(res,event,data)
char             *res;      /* pointer to resource name */
enum event_type event;
char            *data;
{
int _index;
int _index1;
struct resource *_pt;
struct information *_li;
struct request *_lr;
struct answer *_la;
struct transition *_lt;
struct shortcut *_lo;

   /* lockactivity : messages trapped except information messages */
   if ( lockactivity == NO && event != INFORMATION )
	return(0);

/* look for first free index in table */
   _index = NOT_AN_INDEX;
   if ( strcmp(res,MASTER_TOKEN) == 0 )
   {
        _index = MASTER_INDEX;
        if ( (new + _index)->name[0] != '\0' )
        {
             PERROR_B
             "Lock protocol : mastertoken already in message"
             PERROR_E;
             return(-1);
        }
   }
   else 
   {        
        for(_index=FIRST_RESOURCE;_index<RESOURCE_COUNT;_index++)
        {
            if ( (new +_index)->name[0] == '\0' )
                   break;
            /* is resource already listed in message ? */
            else if(strcmp(res,(new + _index)->name) == 0 )
            {
                 PERROR_B
                 "Lock protocol : resource already in message"
                 PERROR_E;
                 return(-1);
            }
        }
   }
           
   if ( _index == NOT_AN_INDEX )
   {
        PERROR_B
        "Too much resources managed by lock protocol"
        PERROR_E;
        return(-1);
   }

/* mark to MARKED the resource list element */
   _index1 = lib_index_of_resource(rsd,res);
   if ( _index1 != NOT_AN_INDEX )
        rsd[_index1].flag = MARKED;
        
/* store resource into new element */      
   _pt = new + _index;
   strncpy(_pt->name,res,sizeof(_pt->name));
   _pt->evttype = event;
   switch(event)
       {
       case INFORMATION :
                _li = (struct information *) &_pt->value1;
                memcpy(_li,data,sizeof(struct information));
                break;
       case REQUEST     :
                _lr = (struct request *) &_pt->value1;
                memcpy(_lr,data,sizeof(struct request));
                break;
       case ANSWER      :
                _la = (struct answer *) &_pt->value1;
                memcpy(_la,data,sizeof(struct answer));
                break;
       case TRANSITION  :
                _lt = (struct transition *) &_pt->value1;
                memcpy(_lt,data,sizeof(struct transition));
                break;
       case SHORTCUT  :
                _lo = (struct shortcut *) &_pt->value1;
                memcpy(_lo,data,sizeof(struct shortcut));
                break;
       }

   /* incrementes broadcast flag */
   *broadcastflag += 1;   
   return(0);
}


/* int add_request_to_rat()
 * Add a request element to a rat queue
 *
 * Argument : pointer to rat pointer ( to link to existing queue )
 *            pointer to request structure
 *            delay value
 *	      pt to resource name for lock message
 *
 * Return  : 0 ok
 *           -1 fatal error
 */
int add_request_to_rat(prat,r,del,res)
struct rat **prat;
struct request *r;
enum delay del;
char      *res;
{
struct rat *_p;
int         _i;

    _p = (struct rat *)malloc(sizeof(struct rat));
    if(!_p)
    {
       PERROR_B
       "Memory allocation for rat %m"
       PERROR_E;
       return(-1);
    }
/* initialise the new element */
    _p->next      = NULL;
    if(*prat)
         _p->previous = *prat;
    else _p->previous = NULL;
    _p->initiator = r->ri;
    lib_timer_set(&(_p->timer),locktimeout);
    _p->generation = r->gn;
    _p->flag      = del;
    _p->sent      = NOTSENT;	/* request not sent yet */
    _p->identlist = NULL;	/* identifier list is NULL by default */
    for(_i=0;_i<MEMBER_COUNT;_i++)
        _p->la[_i] = NOANSWER;
   _p->lr        = r->lr;

/* link it to the list */
    if (*prat)
         (*prat)->next = _p;  
    else *prat = _p;

/* lock protocol message */
    protocol_trace(REQUEST,(int)r->lr,res,r->ri);

    return(0);
}

/* int add_request_to_wish_list()
 * Add a request to wish_list. 
 * If a request with same member initiator is
 * already in wish_list, this request is rejected.(1)
 * 
 * If the new request is a remote request:
 *      - check if a local request is processing and manage conflict
 *              ( lowest initiator is served ) (2)
 * If the new request is a local request:
 *      - delay the local request in case of remote request already processing
 *        (3)
 *      - reject the local request in case of local request already processing
 *        only if the requests are different (1)
 *
 * Argument : pointer to resource name
 *            pointer to request structure
 *            pointer to request pointer to return the new request address.
 *
 * Return   :  0 added or rejected : *currentpt = NULL rejected
 *            -1 fatal error
 *              
 */
int add_request_to_wish_list(res,r,currentpt)
char              *res;
struct request      *r;
struct rat   **currentpt;
{
struct wish_list *_p;		
struct wish_list *_pn;	/* address of the preceding pointer */
struct rat *_q;
struct rat *_qn;
int _inputflag;
int _localprocessing;    /* indicator : local request is processing */
struct rat *_local;      /* pointer to the local processing request */
int _remoteprocessing;   /* indicator : remote request is processing */
int _index;              /* index in resource table */
int _ret;                /* returned value */
//syslog(LOG_ERR,"add_request_to_wish_list has been called");
/* init */
    *currentpt = NULL;
    if (r->ri == localindex)
         _inputflag = LOCAL;          /* new request is LOCAL */
    else _inputflag = REMOTE;         /* new request is REMOTE */
    _localprocessing = 0;             /* init local values */
    _remoteprocessing = 0;

/* control */
    switch(r->lr)
    {
    case DECLARING :
    case ERASING   :
    case RESERVING :
    case MOUNTING  :
    case FREEING   :
    case FAULTING  :
    case GETTING   :
         break;
    default :  
         sprintf(messagestring,
                 "add_request_to_wish_list(), invalid request %d",r->lr);
         PWARNING_B
         messagestring
         PWARNING_E;
         return(-1);
    }

/* string pt control */
    if (res == NULL)
	return(0);

/* generation number control : is it a new request according to local ? */
    _index = lib_index_of_resource(rsd,res);
    if ( _index != NOT_AN_INDEX )
         if ( (rsd+_index)->generation < r->gn )
              return(0);                                /* NO : rejected */
/* for all elements in wish_list */
    _q = NULL;
    _qn = NULL;
    _p  = wishroot;
    _pn = _p;
    while (_p)             /* look if requested resource already listed */
    {           
        if ( strcmp(_p->name,res) == 0)    /* resource found in wish_list */
        {      
             _q = _p->queue;
             _qn = _q;

			 /* we must get out of the loop now if we want 
			  * to keep the _p under consideration
			  */
              break;
        }
	_pn = _p;
	_p = _p->next;
    }

/* if resource not in wish_list */
    if( _qn == NULL )
    {
    /* now pn has the address of the last pointer to wish_list	*/
    /* append new wish at the end of current list		*/
        _p = (struct wish_list *) malloc(sizeof(struct wish_list));
        if (!_p)
        {
            PERROR_B
            "Memory allocation for wish_list %m"
            PERROR_E;
            return(-1);
        }
    /* create wish_list element	*/
        _p->next  = NULL;		
        if ( _pn == NULL )
        {
             _p->previous = NULL;
             wishroot = _p;
        }
        else 
	{
             _p->previous = _pn;
             _pn->next = _p;
        }
		memset (_p->name, '\0', sizeof (_p->name));
        strncpy(_p->name,res,sizeof(_p->name));
        _p->queue = NULL;

    /* increment wishlistcount */
        wishlistcount++;

    /* create rat element */
        _ret = add_request_to_rat(&(_p->queue),r,NOTDELAYED,res);
    /* return address of new rat element in current */
        *currentpt = _p->queue;
        return(_ret);       
    }

/* if resource found in wish_list */
    while(_q)            /* look if existing rat queue elements */
    {
        if ( r->ri == _q->initiator )
	{
	    /* identical local request also queued */
	    if ( _inputflag == LOCAL && _q->lr == r->lr )
            {
/*	         	protocol_trace(REMOTE,res);*/
				protocol_trace (REQUEST, (int) r->lr, res, r->ri);
		 		*currentpt = _q;      /* return queue element pointer */
                 return(0);            /* request not rejected */
	    }
	    else /* other rejected ( case (1) ) : trace conflict */
	    {
	 	 if ( _inputflag == REMOTE )
		      protocol_trace_conflict(REMOTE,res);
		 else protocol_trace_conflict(LOCAL,res);
		 return(0);
	    }
	}
        if ( _q->initiator == localindex )
        {
            _localprocessing++;    /* local already processing   */
            _local = _q;           /* save the rat pointer       */
        }
        else _remoteprocessing++;  /* remote already processing  */
        _qn = _q;                  /* save q address             */
        _q = _q->next;             /* next element               */
    }                        

/* lock exclusion */
/* add the request to the rat queue if possible */
    if ( (_localprocessing != 0) && (_inputflag == REMOTE) &&
         (_remoteprocessing == 0))
    {                              /* conflict */
         if (r->ri < localindex )  /* remote has the priority */
         {                         /* case (2) */
             _ret = add_request_to_rat(&(_qn),r,NOTDELAYED,res);
             *currentpt = _qn->next;
	     /* lock and trace notification */
	     protocol_trace_conflict(LOCAL,res);
	     lock_notification(_local,_p->name,LOCK_ERROR);
	     /* purge rat queue */
             remove_request_from_rat(_p,_local);       
             return(_ret);
         }
         else                       /* remote rejected : case (2) */
	 {                          /* trace conflict */
	     protocol_trace_conflict(REMOTE,res);
             return(0);             
	 }
    }
    /* local is delayed : case (3) */
    else if ( (_remoteprocessing != 0 ) && ( _inputflag == LOCAL))
         _ret = add_request_to_rat(&(_qn),r,DELAYED,res);
    else _ret = add_request_to_rat(&(_qn),r,NOTDELAYED,res);
    /* return address of new rat element in currentpt */
    *currentpt = _qn->next;
    return(_ret);
}

/* int add_answer_to_wish_list()
 * Add an answer to wish_list. 
 * If the ri field (requester index) of the answer match with a local 
 * request initiator field then this 'la' table is updated with the 'la' field
 * at index 'ai' ( answerer index )
 *
 * Argument : pointer to resource name
 *            pointer to answer structure
 *            pointer to rat pointer structure
 *
 * Return   :  0  ok : *currentpt != NULL if request updated
 *             -1 fatal error
 */
int add_answer_to_wish_list(res,a,currentpt)
char              *res;
struct answer       *a;
struct rat **currentpt;
{
struct wish_list *_p;		
struct wish_list **_pp;	/* address of the preceding pointer */
struct rat *_q;

//syslog(LOG_ERR,"add_answer_to_wish_list has been called");
/* init */
    *currentpt = NULL;
    
/* control */
    switch(a->la)
    {
    case OK :
    case NONOK :
         break;
    default :
         sprintf(messagestring,
                 "add_answer_to_wish_list(), invalid answer %d",a->la);
                 PWARNING_B
                 messagestring
                 PWARNING_E;
                 return(-1);
    }

/* only answer for local request are processed */
    if (res && (a->ri == localindex)) 
    {
        _pp = &wishroot;
        _p  = *_pp;
        while (_p)	/* for all elements in wish_list */
        {           /* look if requested resource already listed */
            if ( strcmp(_p->name,res) == 0)
            {      
                 _q = _p->queue;
                 while(_q)
                 {
                     if ( a->ri == _q->initiator )
                     { /* add answer to answer table */
                          _q->la[a->ai] = a->la;
                          *currentpt = _q;
                          /* lock protocol message */
		          protocol_trace(ANSWER,
				         (int)a->la,
					 res,
					 a->ai);
                          break;
                     }
                     _q = _q->next;
                 }
                 break;
            }
	    _pp = &(_p->next);
	    _p  = *_pp;
        }
    }
    return(0);
}

/* int add_transition_to_wish_list()
 * Add a transition to wish_list. 
 *
 * Argument : pointer to resource name
 *            pointer to transition structure
 *            pointer to rat structure pointer
 *
 * Return   :  0 ok : if wish_list updated *currentpt != NULL
 *            -1 fatal error
 */
int add_transition_to_wish_list(res,t,currentpt)
char              *res;
struct transition   *t;
struct rat **currentpt;
{
struct wish_list *_p;		
struct wish_list **_pp;	/* address of the preceding pointer */
struct rat *_q;
//syslog(LOG_ERR,"add_transition_to_wish_list has been called");
 
/* init */
    *currentpt = NULL;
    
/* control */
    switch(t->lt)
        {
        case NOTRANS :
        case DECLARE :
        case ERASE   :
        case RESERVE :
        case MOUNT   :
        case FREET   :
        case FAULTT  :
        case GETT    :
                       break;
        default :      sprintf(messagestring,
                "add_transition_to_wish_list(), invalid transition %d",t->lt);
                        PWARNING_B
                        messagestring
                        PWARNING_E;
                        return(-1);
        }

/* look into wish_list */
    if (res)
        {
        _pp = &wishroot;
        _p  = *_pp;
        while (_p)	/* for all elements in wish_list */
            {           /* look if requested resource already listed */
            if ( strcmp(_p->name,res) == 0)
                 {      
                 _q = _p->queue;
                 while(_q)
                     {
                     if ( t->ri == _q->initiator )
                          {               
                          _q->lt = t->lt;  /* update transition informations */
                          _q->generation = t->gn;
                          *currentpt = _q;   /* save q address into currentpt */
                          /* lock protocol message */
			  protocol_trace(TRANSITION,(int)t->lt,res,t->ri);
                          break;
                          }
                     _q = _q->next;
                     }
                 break;
                 }
	    _pp = &(_p->next);
	    _p  = *_pp;
            }
    }
    return(0);
}

/* void remove_request_from_rat()
 * remove a request from the rat queue. 
 *
 * Argument : pointer to wish_list element
 *            pointer to rat structure
 *
 * Return   : none
 */
void remove_request_from_rat(w,q)
struct wish_list *w;
struct rat *q;
{
struct identifier *_i;
struct identifier *_in;

    if (q)
    {
        if ( q->next != NULL )
             q->next->previous = q->previous;
        if ( q->previous != NULL )
             q->previous->next = q->next;
        else w->queue = q->next;
	_i = q->identlist;
	while(_i)
	{
	    _in = _i->next;
	    free(_i);
	    _i = _in;
	}	
        free(q);     
    }
}

/* struct rat *purge_wish_list()
 * Purge the rat queue associated with r.
 * If LOCAL parameter : purge local requests
 * If REMOTE parameter : purge all remote requests. If one local request is
 * delayed, return local request pointer. this local request has to be 
 * processed.
 * If the rat queue is empty, then remove the wish_list element.
 *
 * Argument : pointer to resource name
 *            parameter either LOCAL or REMOTE
 *
 * Return   : NULL or pointer to rat structure
 *
 * 18-07-2000 Cedric Milesi
 * I did a few change in this routine in order to make sure 
 * that no pointer to stalled data in memory (freed area) is manipulated/used   
 */
struct rat *purge_wish_list(r,param)
char *r;
int param;
{
		struct rat *_ret;
		struct wish_list *_p;
		struct wish_list **_pp;
		struct rat      *_q;
		struct rat		*_next;
//syslog(LOG_ERR,"purge_wish_list has been called");

		_ret = NULL;

		if (r)
		{
				_pp = &wishroot;
				_p = *_pp;
				while(_p)
				{
						if ( strcmp(_p->name,r) == 0 )
						{
								for ( _q = _p->queue; _q; _q = _next )
								{
										/* (CM) we must record the pointer to the next element 
										 * *BEFORE* doing whatever operation over the current 
										 * element (otherwise in case of purge the pointer 
										 * points to freed memory area and it leads to problem when 
										 * we try to access the next element) 
										 */
										_next = _q->next ;

										if ( (param == LOCAL) && (_q->initiator == localindex))
										remove_request_from_rat(_p,_q);
										if ( (param == REMOTE) && (_q->initiator == localindex))
												if (_ret) _ret = _q;    /* save local delayed request */ /* (CM) if not already one saved !!!*/
										/*(CM) I've got a question here: is there always one REMOTE rat per whishlist?) the code above
										 *assume that we can have more than one 
										 */

										if ( (param == REMOTE) && (_q->initiator != localindex))
												remove_request_from_rat(_p,_q);
										
								}
								if ( _p->queue == NULL )        /* no more rat element */
								{
										if ( _p->next != NULL )
												_p->next->previous = _p->previous;
										if ( _p->previous != NULL )
												_p->previous->next = _p->next;
										else 
										{
												wishroot = _p->next;
										}
										free(_p);     
										/* decrement wishlistcount */
										wishlistcount--;
								}     
								break;
						}
						_pp = &(_p->next);
						_p  = *_pp;
				}        
		}
		return(_ret);
}


/* void resource_transition()
 * Activates the eventscript and send a message.
 *
 */
void resource_transition(rn,tr,os,ns)
char *rn;       /* resource name */
int tr;         /* transition requester */
enum resource_status os; /* old status */
enum resource_status ns; /* new status */
{
struct eventscript_data evt; /* data used to call lib_eventscript */

	evt.type = EVT_RESOURCE; 
	evt.display = YES;
	evt.script = scriptactivity;
	evt.index = tr;
	evt.group = grp;
	strcpy(evt.data.r.name,rn);
	evt.data.r.old = os;
	evt.data.r.new = ns;
	lib_eventscript(&evt);     
}

/* int add_to_wait_list()
 * Add an element to the wait_list queue. This is used to store the pid and
 * identifier of each waiting application in order to nofity them in case of 
 * lock on the resource.
 *
 * Argument : application pid
 *	      request identifier
 *            wait operation
 *	      resource name
 *
 * Return   : 0 ok / -1 fatal error
 */
int add_to_wait_list(pid,identifier,wait,res)
pid_t              pid; /* requester pid */
int         identifier; /* requester identifier */
enum wait_operation wait; /* wait operation */
char              *res; /* resource name */
{
struct wait_list *p;
struct wait_list *pn;

   if ( wait == WAIT_FOR_MOUNT || wait == WAIT_FOR_RESERVE )
   {
        p = waitroot;
        pn = p;
        while(p)
        {
            if ( p->pid == pid )
                 return(0);
            pn = p;
            p = p->next;
        }
        p = (struct wait_list *) malloc(sizeof(struct wait_list));
        if (!p)
        {
            PERROR_B
            "Memory allocation for wait_list %m"
            PERROR_E;
            return(-1);
        }
        /* create wait_list element  */
        p->next  = NULL;
        if ( pn == NULL )
        {
            p->previous = NULL;
            waitroot = p;
        }
        else 
        {
            p->previous = pn;
            pn->next = p;
        }
        p->pid = pid;
		p->wait = wait;
		p->identifier = identifier;
		strcpy(p->name,res);
        return(0);                                                   
   }
   return(0);
}

/* int update_resource_count()
 * Update the variable resourcecount by examining the resource table.
 * 
 * Argument : None
 *
 * Return   : Resource count
 */
int update_resource_count()
{
int _r;
int _index;

    _r = 0;
    for(_index=FIRST_RESOURCE;_index<RESOURCE_COUNT;_index++)
    {
         if ( (rsd + _index)->name[0] != '\0' )
               _r++;
    }
    return(_r);
}

/* int update_resourcelist()
 * Update a resource table element. In case of change of the status, activate
 * the RESOURCE eventscript.
 *
 * Argument :   pointer to resource name
 *              status value
 *              owner index
 *              generation number
 *
 * Return   :    0 ok
 *              -1 fatal error
 */
int update_resourcelist(res,sta,own,gen)
char                *res;        /* resource name pointer */
enum resource_status sta;        /* status */
int                  own;        /* owner index */
int                  gen;        /* generation number */
{
int _index;
int _script;
struct resource_def *_s;
enum resource_status _previous;

/* init */
   _script = 0;

/* look for first free or current index in table */
   if ( (_index = find_index_for_resource(res)) == NOT_AN_INDEX )
        return(-1);

/*
/* update resource element */
    _s = rsd + _index;
    _previous = _s->status;
    _script++;
    lib_timer_reset(&(_s->delay));
    if ( sta != NOTKNOWN )
    {
         strncpy(_s->name,res,sizeof(_s->name));
         _s->status = sta;
         _s->generation = gen;
         _s->initiator = own;
         _s->clean = CLEANED;
    }
    else
    {
         memset(_s->name, '\0', sizeof(_s->name));
         _s->status = NOTKNOWN;
         _s->initiator     = NOT_AN_OWNER;
         _s->generation = 0;
    }
   if (_script)          /* status changed : script activation */
   {
        resource_transition(res,own,_previous,_s->status);
   }
        
   /* update resourcecount */
   (*resourcecountpt) = update_resource_count();
   return(0);
}

/* int request_input()
 * Process a request input :
 *       - add the request parameters to the wish_list
 *       - if whislist updated 
 *            - build answer 
 *            - insert answer in new message
 *
 * Argument    : pointer to resource name
 *               pointer to request structure
 *
 * Return      : 0 ok
 *               -1 fatal error
 */
int request_input(res,lr)
char           *res;
struct request *lr; /* used to read request input */
{
struct answer     _a; /* used to build an answer message */
struct rat *_current;
/*int _index;*/

     _current = NULL;
     if (add_request_to_wish_list(res,lr,&_current)==-1)
          return(-1);
     if (_current)
          {
          _a.ri = _current->initiator;
          _a.ai = localindex;
          _a.la = build_answer(res,_current->lr);
/*          if ( (_index = lib_index_of_resource(rsd,res)) != NOT_AN_INDEX ) */
/*               (rsd+_index)->clean = CLEANED; */
          /* lock protocol message */
	  protocol_trace(ANSWER,(int)_a.la,res,_a.ai);          
          return(insert_resource_in_newlist(res,ANSWER,(char *)&_a));
          }
     return(0);
}

/* int answer_input()
 * Process an answer :
 *       - add the answer parameters to the wish_list
 *       - if whislist updated 
 *            - process lock protocol
 *            - if WDXLOCK : local transition
 *                  - build a transition 
 *                  - insert transition in new message
 *                  - update resource list
 *                  - purge wish_list from local request
 *		    - notification of the initiator
 *            - if WDXUNLOCK : local transition refused.
 *                  - purge wish_list from local request
 *		    - notification of the initiator
 *
 * Argument    : pointer to resource name
 *               pointer to answer structure
 *
 * Return      : 0 ok
 *               -1 fatal error
 */
int answer_input(res,la)
char          *res;
struct answer  *la; /* used to read answer input */
{
struct transition    _t; /* used to build a transition message */
struct rat    *_current;
enum lock_protocol _lock;

     _current = NULL;
     /* add answer to wish list */
     if (add_answer_to_wish_list(res,la,&_current)==-1)
          return(-1);
     if ( _current )
     {
         _lock = lock_protocol(_current);
         if ( _lock == WDXLOCK)
         {
	      /* transition build */
              _t.lt = build_transition(_current->lr);
              _t.ri = _current->initiator; /* localindex */
              _current->lt = _t.lt;
              _current->generation = lib_build_generation();
              _t.gn = _current->generation;

              /* lock protocol message and notification */
	      protocol_trace(TRANSITION,(int)_t.lt,res,_t.ri);
	      lock_notification(_current,res,LOCK_SUCCESS);

	      /* resource list update */
              if (update_resourcelist(res,
                                      build_status(_current->lt),
                                      _current->initiator,
                                      _current->generation) == -1)
                   return(-1);
	      /* newlist update */
              if (insert_resource_in_newlist(res,
                                             TRANSITION,
                                             (char *)&_t) == -1)
                   return(-1);
	      /* wish list purge */
              purge_wish_list(res,LOCAL);
         }
         if ( _lock == WDXUNLOCK )
         {
	      /* NOTRANS message */
              _t.lt = NOTRANS;
              _t.ri = _current->initiator; /* localindex */
              _t.gn = 0;

              /* lock protocol message and notification */
	      protocol_trace_denied(DENY_MESSAGE,_current->lr,
                                    res,_current->initiator);
	      lock_notification(_current,res,LOCK_ERROR);

	      /* wish list purge and newlist update */
              purge_wish_list(res,LOCAL);
              if (insert_resource_in_newlist(res,
                                             TRANSITION,
                                             (char *)&_t) == -1)
                  return(-1);
         }
    }
   return(0); 
}

/* int transition_input()
 * Process a transition input :
 *       - if the transition is NOTRANS, then purge the wish_list
 *       - else
 *       -      add the transition parameters to the wish_list
 *       -      if whislist updated 
 *                      - update resourcelist.
 *                      - purge wish_list from remote request.
 *                      - activate local delayed request if exist
 *
 * Argument    : pointer to resource name
 *               pointer to transition structure
 *
 * Return      : 0 ok
 *               -1 fatal error
 */
int transition_input(res,lt)
char             *res; /* resource name */
struct transition *lt; /* lock transition pointer */
{
struct request  _re; /* used to build request message for delayed local request*/
struct rat *_current;
   
   _current = NULL;
   /* add transition to wish list */
   if (add_transition_to_wish_list(res,lt,&_current)==-1)
        return(-1);
   if (_current)
   {
	/* lock protocol trace */
        if ( lt->lt != NOTRANS )
        {
		/* update resource list */
                if (update_resourcelist(res,
                                        build_status(_current->lt),
                                        _current->initiator,
                                        _current->generation) == -1)
                     return(-1);
        }
	/* purge wish list and if local request pending : send request */
        if ((_current=purge_wish_list(res,REMOTE)))
        {
	     if ( _current->sent == NOTSENT )
	     {
		  _current->flag = NOTDELAYED;
		  _current->sent = SENT;
                  _current->la[localindex] = build_answer_to_local_request(res,
                                                                  _current->lr);
                  _re.lr = _current->lr;
                  _re.ri = _current->initiator;
                  _re.gn = _current->generation;
                  return(insert_resource_in_newlist(res,REQUEST,(char *)&_re));
	     }
        }
   }
   return(0);
}            

/* int shortcut_input()
 * Process a shortcut input :
 *       - update the resource list
 *       - trace the event
 *
 * Argument    : pointer to resource name
 *               pointer to shortcut structure
 *
 * Return      : 0 ok
 *               -1 fatal error
 */
int shortcut_input(res,lo)
char             *res; /* resource name */
struct shortcut *lo;   /* lock shortcut pointer */
{
	/* update resource list */
	protocol_trace(SHORTCUT,(int)lo->status,res,lo->initiator);
        update_resourcelist(res,
                            build_status(lo->status),
                            lo->initiator,
                            lo->generation);
        return(0);
}            

/* int information_input()
 * Process an information input :
 *              - update resource list.
 *
 * Argument    : pointer to resource name
 *               pointer to information structure
 *
 * Return      : 0 ok
 *               -1 fatal error
 */
int information_input(res,li)
char *res;
struct information *li;
{
int _index;
int _script;   /* script flag */
int _conflict; /* conflict flag */
int _mismatch; /* mismatch flag */
struct resource_def *_s;
enum resource_status _previous;

/* init */
   _script = 0;
   _conflict = 0;
   
/* look for first free or current index in table */
   if ( (_index = find_index_for_resource(res)) == NOT_AN_INDEX )
        return(-1);

/* update resource element */
   _s = rsd + _index;
   _previous = _s->status;
   (rsd+_index)->clean = CLEANED;
   switch(li->status)
        {
        case FREE :     /* No conflict : requester status are preserved */
             switch(_s->status)
                  {
                  case NOTKNOWN :
                       _script++;
                       break;
                  default :
                       break;
                  }
             break;
        case NOTKNOWN : /* No conflict : requester status are preserved */
             switch(_s->status)
                  {
                  case FREE :
                       _script++;
                       break;
                  default :
                       break;
                  }
             break;
        case RESERVED : /* Conflict with other requesters */
             switch(_s->status)
                  {
                  case NOTKNOWN :
                  case FREE :
                       _script++;
                       break;
                  case RESERVED :
                       if ( _s->initiator != li->initiator )
                            _conflict++; /* conflict */
                       break;
                  default : /* conflict */
                       _conflict++;
                       break;
                  }
             break;
        case MOUNTED : /* Conflict with other requesters */
             switch(_s->status)
                  {
                  case NOTKNOWN :
                  case FREE :
                       _script++;
                       break;
                  case MOUNTED :
                       if ( _s->initiator != li->initiator )
                            _conflict++; /* conflict */
                       break;
                  default : /* conflict */
                       _conflict++;
                       break;
                  }
             break;
        case FAULT : /* Conflict with other requesters */
             switch(_s->status)
                  {
                  case NOTKNOWN :
                  case FREE :
                       _script++;
                       break;
                  case FAULT :
                       if ( _s->initiator != li->initiator )
                            _conflict++; /* conflict */
                       break;
                  default : /* conflict */
                       _conflict++;
                       break;
                  }
             break;
        }
   /* conflict */
   _mismatch = 0;	/* No mismatch */
   if (_conflict)
   {
/*       sprintf(messagestring,"recorded generation: %d",li->generation);
 *      PERROR_B
 *      messagestring
 *      PERROR_E;
 *      sprintf(messagestring,"local generation: %d",_s->generation);
 *      PERROR_B
 *      messagestring
 *      PERROR_E;
*/      
       if ( li->generation > _s->generation )
            _script++;                                   /* keep fresher */
       else
       { 
	    if ( (li->generation == _s->generation) &&      
                 ( li->initiator < _s->initiator ))    /* keep lowest index */
                      _script++;
            else _mismatch++;
       }
   }
   /* mismatch */
   if ( _mismatch )
        mismatch_mark(res,li->initiator);

   /* script */    
   if (_script)
       {
       strncpy(_s->name,res,sizeof(_s->name));
       _s->status = li->status;
       _s->generation = li->generation;
       _s->initiator = li->initiator;
       resource_transition(res,li->initiator,_previous,_s->status);
       }

   /* update resourcecount */
   (*resourcecountpt) = update_resource_count();
   return(0);
}

/* void set_not_cleaned()
 * Set clean field for all resources to NOT_CLEANED
 *
 * Argument : None
 *
 * Return : None
 */
void set_not_cleaned()
{
int _x;

   for ( _x=0;_x<RESOURCE_COUNT;_x++)
         (rsd+_x)->clean = NOT_CLEANED;
}

/* void set_cleaned()
 * Set clean field to CLEANED for resource 'name' if locked by 'id
 *
 * Argument : id   : requester index
 *            name : pt to resource name
 *
 * Return : None
 */
void set_cleaned(id,name)
int  id;
char *name;
{
int _index;
   if ( (_index = lib_index_of_resource(rsd,name)) != NOT_AN_INDEX )
   {
        if ( (rsd+_index)->initiator == id )
        {
             switch((rsd+_index)->status)
             {
             case RESERVED :
             case FAULT :
             case MOUNTED :
                  (rsd+_index)->clean = CLEANED;
                  break;
             default :
                  break;
             }
        }
   }
}

/* void reset_not_cleaned()
 * Test all resources locked by 'id'. If not CLEANED then force them to FREE
 * state.
 *
 * Argument : id of initiator
 *
 * Return : None
 */
void reset_not_cleaned(id)
int id;
{
int _x;
   for ( _x=0;_x<RESOURCE_COUNT;_x++)
   {
        if ( (rsd+_x)->initiator == id )
        {
             switch((rsd+_x)->status)
             {
             case RESERVED :
             case FAULT :
             case MOUNTED :
                  if ( (rsd+_x)->clean == NOT_CLEANED )
                  {
                        resource_transition((rsd+_x)->name,id,
                                         (rsd+_x)->status,FREE);
                        (rsd+_x)->status = FREE;
                        (rsd+_x)->initiator = id;
                  }
                  break;
             default :
                  break;
             }
        }
   }
}

/* void validate_delayed_event()
 * Look in the resource list for resource marked. ( after masterholder failure )
 * decrement counter and if timer evetn set trigger script.
 *
 * Arguement : decrement value
 */
void validate_delayed_event(int dec)
{
int _i;
struct eventscript_data evt; /* data used to call lib_eventscript */

    for(_i=0;_i<RESOURCE_COUNT;_i++)
    {
	if ( (rsd+_i)->name[0] != '\0' && lib_timer_alive(&((rsd+_i)->delay)) )
	{
	    lib_timer_decrease(&(rsd+_i)->delay,dec);
	    if ( lib_timer_event(&(rsd+_i)->delay) )
	    {

	         lib_timer_reset(&((rsd+_i)->delay));
	         evt.type = EVT_RESOURCE; 
	         evt.display = NO;
	         evt.script = scriptactivity;
	         evt.index = (rsd+_i)->oldinit;
	         evt.group = grp;
	         strcpy(evt.data.r.name,(rsd+_i)->name);
		 evt.data.r.old = (rsd+_i)->oldstatus;
		 evt.data.r.new = FREE;
	         lib_eventscript(&evt);     
	    }
	}
    }
}

/* int validate_wait_list()
 * Look into the wait list to check if a resource is mounted or reserved. If
 * such a case notify the requesting process.

 * Argument : None
 *
 */
int validate_wait_list()
{
int _ret;
int _own;
enum resource_status _status;
char _messuc[LEN+RESOURCE_LEN+2];
struct wait_list *_p;
struct wait_list *_pn;
int _f; /* flag */
    _ret = 0;
    _p = waitroot;
    while(_p)                    /* for all elements in wait_list */
	{
//syslog(LOG_ERR,"One more loop _p 0x%x waitroot 0x%x ",_p,waitroot);
			_f = 0;
			_status = lib_status_of_resource(rsd,_p->name);
			switch(_status)
			{
					case RESERVED :
							if ( _p->wait == WAIT_FOR_RESERVE )
									_f++;
							break;
					case MOUNTED :
							if ( _p->wait == WAIT_FOR_MOUNT )
									_f++;
					default :
							break;
			}
			if(_f)
			{
					_own = lib_requester_of_resource(rsd,_p->name);
					sprintf(_messuc,"%s:%s",(mbr+_own)->name,_p->name);		
					ker_build_notification(_p->pid,
									PROCESS_SUCCESS,
									_p->identifier,
									0,
									_status,
									_messuc);
			}
			_pn = _p;
			_p = _p->next;
			if(_f)
			{
					if ( _pn->next != NULL )
							_pn->next->previous = _pn->previous;
					if ( _pn->previous != NULL )
							_pn->previous->next = _pn->next;
					else waitroot = _pn->next;
					free(_pn);
			}
	}
	return(0);
}

/* int validate_wish_list()
 * Decrements validity field of any wish_list element. 
 * If validity reaches LOCK_TIMEOUT and request is LOCAL then process the 
 * lock protocol
 * If validity reaches TRANSITION_TIMEOUT and request is REMOTE then purge
 * the wish_list. Activate the LOCAL DELAYED request.
 *
 * Argument : decrementation value
 *            pointer to broadcastflag
 *
 * (18-07-2000) Cedric Milesi 
 * I did a few changes to this code (see comments in purge_wish_list). This is *REALLY* important
 * that in each case where we are not sure of the state of a memory area 
 * we don't use data from it.
 */
int validate_wish_list(decrement,brdflgpt)
int decrement;	  /* decrementation value */
int *brdflgpt;    /* pointer to broadcast flag */
{
struct wish_list *_p;	/* pt to wish list */
struct wish_list *_previous,*_next; /* pt to wish list */
char   res_name[256];     
struct wish_list *_pn;	/* pt to wish list */	
struct rat *_q;         /* pt to rat list */
struct rat *_current;   /* pt to rat list */
struct rat *_nextq;
struct transition    _t; /* used to build a transition message */
struct request _re;
enum lock_protocol _lock;
int  _ret;
char        string_tmp[1024];
    _ret = 0;
    broadcastflag = brdflgpt;
    for(_p=wishroot;_p;_p = _next)
    {
			_previous = _p->previous;
			_next	=	_p->next;
			strncpy(res_name,_p->name,256 );

        for(_q=_p->queue; _q; _q = _nextq )
		{
			_nextq = _q->next ;
            if ( _q->flag == NOTDELAYED )
            {
                 lib_timer_decrease(&(_q->timer),decrement);
                 /* LOCAL request and LOCK_TIMEOUT reached */
                 if ( (_q->initiator == localindex) &&
                      (lib_timer_reached(&(_q->timer),LOCK_TIMEOUT)))
                 {
                      _lock = lock_protocol(_q);
                      if ( _lock == WDXLOCK)
                      {
			   /* build transition */
                           _t.lt = build_transition(_q->lr);
                           _t.ri = _q->initiator; /* localindex */
                           _q->lt = _t.lt;
                           _q->generation = lib_build_generation();
						   _t.gn = _q->generation;
			   /* lock notification and trace notification */
						   protocol_trace(TRANSITION,(int)_t.lt,res_name,_t.ri);
						   lock_notification(_q,res_name,LOCK_SUCCESS);
                           _ret = _ret | update_resourcelist(res_name,
                                                           build_status(_q->lt),
                                                           _q->initiator,
                                                           _q->generation);
                           _ret = _ret | insert_resource_in_newlist(res_name,
                                                                  TRANSITION,
                                                                  (char *)&_t);
                      }
					  else
					  {
							  /* lock and trace notification */
							  protocol_trace_denied(DENY_TIMEOUT,_q->lr,
											  res_name,_q->initiator);
							  lock_notification(_q,res_name,LOCK_ERROR);
					  }
					  /*(CM) removed (beg)
					    if ( _p->previous != NULL )
					    		  _pn = _p->previous;
					    else _pn = wishroot;
					     removed end
					  */
					  
					  /* purge wish list */
					  purge_wish_list(res_name,LOCAL);
					  
					  /* (CM) note that we are no more sure that current _p is still valid
					   */
						
					  /*(CM) removed (beg)
					  _p = _pn;
					  (CM) removed (end)
					  */

					  break;
				 }       
                 /* REMOTE request and TRANSITION_TIMEOUT reached */
                 if ( (_q->initiator != localindex) && 
                      (lib_timer_reached(&(_q->timer),TRANSITION_TIMEOUT)))
                 {
						 protocol_trace_denied(DENY_TIMEOUT,_q->lr,
										 _p->name,_q->initiator);
			  	/*(CM) removed (beg)
			    if ( _p->previous != NULL )
			         _pn = _p->previous;
			    else _pn = wishroot;
			    (CM) removed (beg)
				*/

		      /* purge wish list, if local request pending process it */
			  _current = purge_wish_list(res_name,REMOTE);

			  /* 18-07-2000  Update Cedric Milesi 
			     note that we are no more sure that current _p is still valid :
			  */

			  /*(CM) removed (beg)
			    _p = _pn;
			    (CM) removed (end)
              */
			  if (_current)
			  {
					  if ( _current->sent == NOTSENT )
					  {
							  _current->flag = NOTDELAYED;
							  _current->sent = SENT;
							  _re.lr = _current->lr;
							  _re.ri = _current->initiator;
							  _re.gn = _current->generation;
							  _current->la[localindex] = 
									  //						  build_answer_to_local_request(_p->name,
									  //										  _re.lr);
									  build_answer_to_local_request(res_name, _re.lr);

							  //					  protocol_trace(ANSWER,
							  //								  (int)_current->la[localindex],
							  //								  _p->name,
							  //								  localindex);
									  protocol_trace(ANSWER,(int)_current->la[localindex],res_name,localindex);

 
									  //									  _ret = _ret | 
									  //											  insert_resource_in_newlist(_p->name,
									  //															  REQUEST,
									  //                                                                  (char *)&_re);
									  _ret = _ret | insert_resource_in_newlist(res_name,REQUEST,(char *)&_re);

					  }
			  }
					  break;                      
				 }       
			}  
		}
	}
    return(0);
}

/* void protocol_trace()
 * Send a message to the log file to trace the lock protocol events.
 *
 * Arguments : event type
 *	       event value
 *	       resource name
 *	       member index
 *
 * return    : None
 */
void protocol_trace(enum event_type event,int value,char *res,int index)
{
char *pt;
char *f;
char *s;

  if( strcmp(res,mastertoken) == 0 )
       pt = MASTER_TOKEN_STRING;
  else pt = res;
  switch(event)
  {
  case REQUEST :
    f = "Lock request \"%s\" for %s from %s";
    s = lock_request_to_s((enum lock_request)(value));
    break;
  case ANSWER : 
    f = "Lock answer \"%s\" for %s from %s";
    s = lock_answer_to_s((enum lock_answer)(value));
    break;
  case TRANSITION :
    f = "Lock transition \"%s\" for %s from %s";
    s = lock_transition_to_s((enum lock_transition)(value));
    break;
  case SHORTCUT :
    f = "Mandatory lock \"%s\" for %s from %s";
    s = lock_transition_to_s((enum lock_transition)(value));
    break;
  }
  sprintf(messagestring,f,s,pt,(mbr+index)->name);
  PINFOR_B
  messagestring
  PINFOR_E;
}

/* void protocol_trace_denied()
 * Trace a denied lock message
 *
 * Arguments : deny event
 *	       lock request
 *	       pt to resource name
 *	       initiator index
 *
 * Return    : None
 */
void protocol_trace_denied(enum deny_event deny,enum lock_request r,char *res,int index)
{
char *pt;
char *f;

  if( strcmp(res,mastertoken) == 0 )
       pt = MASTER_TOKEN_STRING;
  else pt = res;
  if ( deny == DENY_TIMEOUT )
       f = "Lock request \"%s\" for %s denied by timeout";
  else f = "Lock request \"%s\" for %s denied";
  sprintf(messagestring,f,lock_request_to_s(r),pt);
  PINFOR_B
  messagestring
  PINFOR_E;   
}

/* void protocol_trace_conflict()
 * Trace a conflict lock message
 *
 * Arguments : conflict origin
 *	       pt to resource name
 *
 * Return    : None
 */
void protocol_trace_conflict(int conflict,char *res)
{
char *f;
char *pt;

  if( strcmp(res,mastertoken) == 0 )
       pt = MASTER_TOKEN_STRING;
  else pt = res;
  switch(conflict)
  {
  case LOCAL :
      f = "Wish conflict for resource %s, local request already processing";
      break;
  case REMOTE :
      f = "Wish conflict for resource %s, remote request already processing";
      break;
  }
  sprintf(messagestring,f,pt);
  PINFOR_B
  messagestring
  PINFOR_E;
}

/* void lock_notification()
 * Notify all the application who have asked for the request
 *
 * Argument : pt to rat request
 *	      pt to resource name
 *	      lock status
 *
 * Return   : none
 */
void lock_notification(struct rat *r,char *n,enum lock_status s)
{
enum process_status  _t;	/* process status */
char                *_p;	/* pt to resource name */
char                *_f;	/* pt to message format */
char _m[LEN+RESOURCE_LEN+2];    /* pt to formatted message */
struct identifier   *_i;	/* pt to identifier structure */

   if ( s == LOCK_ERROR )
   {
	_t = api_statuss[r->lr].error;
	_f = api_statuss[r->lr].error_message;
   }
   else 
   {
	_t = api_statuss[r->lr].success;
	_f = api_statuss[r->lr].success_message;
   }

   if ( strcmp(n,MASTER_TOKEN) == 0 )
	_p = MASTER_TOKEN_STRING;
   else _p = n;  

   sprintf(_m,_f,_p);   
   _i = r->identlist;
   while(_i)
   {
	if ( _i->pid == getpid() )
	{
	     if ( _i->identifier == AUTOMATIC_IDENTIFIER )
		  automaticflag = 0; /* reset request flag */
	}
	else
	{
	     ker_build_notification(_i->pid,
	    		            _t,
	  		            _i->identifier,
			            r->generation,
			            lib_status_of_resource(rsd,n),
			            _m);
	}
	_i = _i->next;
   }
}

/* int add_identifier_to_rat()
 * add an identifier structure to rat queue element
 *
 * Argument : pt to rat element
 *	      pid_t value
 *	      identifier value
 *
 * Return   : -1 if error / 0 if OK
 */
int add_identifier_to_rat(struct rat *c,pid_t p,int i)
{
struct identifier *_p;
struct identifier *_v;

    if(c)
    {
	_p = (struct identifier *) malloc(sizeof(struct identifier));
	if (!_p)
	{
	    PERROR_B
	    "Memory allocation for identifier list %m"
	    PERROR_E;
	    return(-1);
	}
	/* setup new element */
	_p->next = NULL;
	_p->pid = p;
	_p->identifier = i;

	/* link it to the list */
	_v = c->identlist;
	if(!_v)
	    c->identlist = _p;
	else
	{
	    while(_v->next)
	        _v = _v->next;
	    _v->next = _p;	    
	}        	    
    }
    return(0);
}

/* int operation_preprocess()
 * When a operation message is received some pre control are performed.
 *
 * Argument : pt to operation
 *	      lock request value
 *	      index of resource in resourcelist
 *	      pt to error returned
 *
 * Return   : 0 all control OK / -1 operation terminated
 */
int operation_preprocess(struct operation *c,enum lock_request t, int x, int *r)
{
enum resource_status _s; /* status of resource */
int                  _o; /* owner of resource */
char                *_p; /* pt to printable resource name */
char _m[LEN+RESOURCE_LEN+2]; /* formatted message */
int                  _i; /* index of resource */

    *r = LOCK_S;
    /* general control on request */
    if ( (*resourcecountpt+wishlistcount) >= (RESOURCE_COUNT-FIRST_RESOURCE))
    {
	if ( x == NOT_AN_INDEX )
	     *r = LOCK_T; /* operation rejected */
    }
    else if ( ( t == MOUNTING ) &&
         ( spc_check_mount_point(c->operationdata,WDX_PROCESS) != 0 ) )
         *r = LOCK_M; /* Not a local mount point */
 
    else if ( ( t == GETTING ) &&
	( lib_status_of_resource(rsd,c->operationdata) == MOUNTED ) )
	*r = LOCK_P;

    if ( *r != LOCK_S )
	 return(-1);
    /* specific control to accept request even if the resource is locally
       locked */
   switch(t)
   {
   case RESERVING :
   case MOUNTING :
   case FAULTING :
   case GETTING :
        _s = lib_status_of_resource(rsd,c->operationdata);
        _o = lib_requester_of_resource(rsd,c->operationdata);
	/* resource already locally locked */
	if ( _s == api_statuss[t].waited && _o == localindex )
	{
	    /* notify the application */
	    if(strcmp(c->operationdata,MASTER_TOKEN ) == 0)	    
		 _p = MASTER_TOKEN_STRING;
	    else _p = c->operationdata;
	    sprintf(_m,api_statuss[t].success_message,_p);
	    _i = find_index_for_resource(c->operationdata);
	    ker_build_notification(c->pid,
				   api_statuss[t].success,
				   c->identifier,
				   (rsd+_i)->generation,
				   _s,
				   _m);
	    *r = LOCK_D;		/* lock request done */
	    return(-1);
	}
        if ( t == GETTING && strcmp(c->operationdata,MASTER_TOKEN) == 0)
        {
             struct shortcut _o;
             struct request _req;
             struct rat *_current;
	     int _ret;
             /* ajout request dans rat */
             _req.lr = t;
             _req.ri = c->index;
             _req.gn = (rsd+x)->generation;
             _ret = add_request_to_wish_list(c->operationdata,&_req,&_current);
             if ( (_ret) || (!_current) )
             {
                  *r = LOCK_R;
                  return(-1);
             }
             _ret = add_identifier_to_rat(_current,c->pid,c->identifier);
             if ( _ret )
             {
                  *r = LOCK_R;
                  return(-1);
             }        
             /* creation of message shortcut */
             /* t is a request : build the associated transition then using
                the transition build the resource status */
             _o.status = build_transition(t);       
             _o.initiator = c->index;
             _o.generation = lib_build_generation();
             if ( insert_resource_in_newlist(c->operationdata,
                                             SHORTCUT,
                                             (char *)&_o) == -1 )
             {
                  *r = LOCK_R;
                  return(-1);
             }        

             /* lock protocol message and notification */
             protocol_trace(SHORTCUT,(int)_o.status,
                            c->operationdata,c->index);
             lock_notification(_current,c->operationdata,LOCK_SUCCESS);

             /* update_resource list */
             if ( update_resourcelist(c->operationdata,
                                       build_status(_o.status),
                                       c->index,
                                       _o.generation) == -1 )             
             {
                  *r = LOCK_R;
                  return(-1);
             }        
             /* purge wish_list LOCAL */      
             purge_wish_list(c->operationdata,LOCAL);
   
             /* result */
             *r = LOCK_D; /* done */
    	     return(-1);
	}
        break;
   default :
	break;	
   }
   return(0);

}

/* void mismatch_mark()
 * A mismatch has been found by the information input routine.
 *
 * Argument : resource name pt
 *	      index of the mismatching member
 *
 * Return   : None
 */
void mismatch_mark(char *r,int i)
{
struct mismatch_list *_m;
struct mismatch_list *_c;
int                   _f;   /* found flag */

    /* search if the mismatch is already registered */
    _m = mismatchroot;
    _c = _m;
    _f = 0;
    while(_m)
    {
	_c = _m;
	if(_m->initiator == i && strcmp(_m->name,r) == 0 )
	{
	    _f++;	
	    break;
        }
        _m = _m->next;
    }
    if(_m && _f)    /* found */
    {
	_m->alive = MARKED;
    }
    else	    /* not found : add to list */
    {
	_m = (struct mismatch_list *) malloc(sizeof(struct mismatch_list));
	if(!_m)
	{
	    PERROR_B
	    "Memory allocation for mismatch_list %m"
	    PERROR_E;
	}
	else
	{
	    _m->next = NULL;
	    strcpy(_m->name,r);
	    _m->initiator = i;
	    _m->alive = MARKED;
	    lib_timer_set(&(_m->timer),locktimeout);
	    if(_c == NULL)
	    {
		mismatchroot = _m;
		_m->previous = NULL;
	    }
	    else 
            {
		_c->next = _m;
		_m->previous = _c;
	    }
	}	
    }
}

/* void mismatch_unmark()
 * Unmark all the mismatch_list element
 *
 * Argument : None
 *
 * Return   : None
 */
void mismatch_unmark()
{
struct mismatch_list *_m;

    _m = mismatchroot;
    while(_m)
    {
	_m->alive = NOT_MARKED;
	_m = _m->next;
    }
}

/* void mismatch_reset()
 * Reset all mismatch element with initiator equal to id if the mismatch element
 * is not MARKED
 *
 * Argument : id of the initiator
 *
 * Return   : None
 *
 * 19-7-2000 Cedric Milesi
 *
 * Modification done in order to never use stalled area of memory.
 *
 */
void mismatch_reset(int i)
{
struct mismatch_list *_m;
struct mismatch_list *_next;
struct mismatch_list *_c;

    /* search if the mismatch is already registered */
    _m = mismatchroot;
    _c = _m;
    while(_m)
    {
	_c = _m; /*( I DON'T SEE what is the use of this variable */

	/* (CM) We get the pointer to the next element *NOW* (and not after
	 * the current element has been freed  
	 */ 
	_next =  _m->next;
	/* end (CM) modification */
	
	if(_m->initiator == i && _m->alive == NOT_MARKED )
	{
	    if(_m->next != NULL)
		 _m->next->previous = _m->previous;
	    if(_m->previous != NULL)
		 _m->previous->next = _m->next;
            else mismatchroot = _m->next;
	    free(_m);
	}
        _m = _next;
    }
}

/* void mismatch_validate()
 * Validate all the mismatch element
 * If timer event set, then trigger the eventscript with MISMATCH type
 *
 * Argument : Decrement counter value
 *
 * Return   : None
 */
void mismatch_validate(int d)
{
struct mismatch_list *_m;
struct eventscript_data evt; /* data used to call lib_eventscript */

    /* search if the mismatch is already registered */
    _m = mismatchroot;
    while(_m)
    {
	lib_timer_decrease(&(_m->timer),d);
	if(lib_timer_event(&(_m->timer)))
	{
	    lib_timer_reset(&(_m->timer));

	    evt.type = EVT_MISMATCH; 
	    evt.display = NO;
	    evt.script = scriptactivity;
	    evt.index = _m->initiator;
	    evt.group = grp;
	    strcpy(evt.data.i.name,_m->name);
	    lib_eventscript(&evt);     
	}
        _m = _m->next;

    }
}
