/*
 *   WDX: WitchDoctor-WDX, versatile distributed lock manager
 *   Copyright (C) 2003 Dominique Chabord,
 *   Please refer to credits section below
 *
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   ************************************************
 *
 *   History and credits:
 *   First version of WitchDoctor-WDX was based on:
 *
 *   DECwatchdog/Autopilot, watchdog and lock manager
 *   Copyright (C) 2003 HP
 *
 *   Developed by
 *     Computer Special Systems, Annecy France
 *     A Division of Digital Equipment Corporation
 *   and
 *     Custom-Systems, Annecy France
 *     An engineering center of Compaq
 *
 *   Provided under GNU-GPL terms by HP, 30 June 2003
 *   GNU-GPL replaces previous license statements
 *
 *   Rebranded, ported and modified for Shaman-X project, 15 July 2003
 *   by Dominique Chabord
 *   and maintained by other contributors as mentionned in the code
 *
 *   ************************************************
 *
 *   WDX is promoted by Shaman-X project
 *
 *   Check for new versions and documentation at
 *   http://www.shaman-x.org
 *
 *   ************************************************
 *
 *   Most of WDX code is designed to run unchanged on
 *   different operating systems, including at V1.0 time:
 *   Gnu-Linux, WindowsNT, DigitalUNIX and OpenVMS.
 *   Operating system specifics are identified by dedicated files.
 *
 *   Evolutions
 *   ----------
 *   July-2003     Dominique Chabord    WitchDoctor-WDX V1.0
 *                 Creation
 *
 *   ************************************************
 *
 *
 *	wdx_watchdog.c
 *
 *	Module to implement the watchdog protocol
 *
 * BE CAREFULL : To initialize this module the wat_init() procedure
 * must be called. It can be done through the 'function_table' structure.
 *
 *      Public Routines :
 *
 *      void wat_init()
 *	void wat_analyze_networks()
 *      int  wat_input()
 *	void wat_activity_set()
 *      void wat_validate_networks()
 *      int  wat_local_member_transition()
 *      int  wat_remote_member_transition()
 *      void wat_failure()
 *      int  wat_node_exception()
 *
 *	Local Routines :
 *
 *	void trigger_script()
 *	void member_transition()
 *	int  how_many_members()
 *	int  is_multi_networks()
 *	int  member_confirmation()
 *      int  input_net()
 *
 ************************************************************************/
/*
 *
 *  18-Feb-94 -- Cornu
 *      Creation
 *
 * MK001 M.Kennealy   Added parameter brdflgpt to wat_local_member_transition()
 *
 */

#undef  WDX_WATCHDOG_MODULE
#define WDX_WATCHDOG_MODULE

#include "wdx.h"

#define SINGLE 1
#define BOTH 2

/* local pointers definitions */
static struct group *gfp;
static struct member *mfp;

/* timer definition */
static int automatictimer;	    /* timer value */
static struct wdx_timer autodelay;  /* automatic delay timer */
static int localpriority;	/* local priority */
static int maxpriority;	        /* max priority value + 1 */
static int conftimer;           /* network protocol confirmation timeout */
/* not a static, because lock_protocol can reset this value */
int automaticflag;	/* flag to start autodelay only once */

/* mask */
static enum mask watchdogactivity;	    /* activity flag */
static enum mask watchdogautomatic;	    /* lock activity flag */

/* watchdog protocol flags */
static int masterowner; 	        /* owner of MASTERTOKEN */
static int there_was_a_substitute; 	/* no SUBSTITUTE */

/*
 * Local routines prototypes 
 */
void trigger_script(char *s,char *a,char *m);
void member_transition(int mi, enum member_status os, enum member_status ns);
int how_many_members();
int is_multi_networks(struct member *m);
int member_confirmation(struct member *m, int i);
int input_net(int n, struct watchdog *w, int i);

/*
 * PUBLIC ROUTINES : prefixe is "wat"
 */

/* void wat_init()
 * Initialise static pointers and counters used by the watchdog_protocol
 * For COLD_INIT or COLD_REINIT : 
 *	- initialize all networks counters.
 *	- initialize all member status.
 * For WARM_REINIT only pointers and static counters are updated
 *
 * Argument : COLD_INIT or COLD_REINIT or WARM_REINIT mode
 *	      pt to group structure
 *	      pt to message ( NOT USED )
 */
void wat_init(s,p,m)
enum startup_mode s;
struct group *p;
struct message *m;
{
int i;		 /* member index */
enum networks j; /* network index */

        gfp         = p;
        mfp         = p->memberlist;
	
        lib_timer_reset(&autodelay);
	automaticflag = 0;
	watchdogactivity = NO;
	watchdogautomatic = NO;

	/* set automatic timer to max of broadcast */
	/* set maxpriority to max of priority */
	automatictimer = 0;
	maxpriority = 0;
	for(i=0;i<gfp->membercount;i++)
	{
	    if((mfp+i)->broadcast > automatictimer)
		automatictimer = (mfp+i)->broadcast;
	    if((mfp+i)->broadcast > conftimer )
		conftimer = (mfp+i)->broadcast;
	    if((mfp+i)->priority > maxpriority)
		maxpriority = (mfp+i)->priority;
	}
	automatictimer++;
	maxpriority++;

	/* set localpriority */
	localpriority = (mfp+localindex)->priority;

	switch(s)
	{
	case COLD_REINIT : /* cold re-init */
	case COLD_INIT :  /* cold init */
	    masterowner = NOT_AN_OWNER; /* no owner */
	    there_was_a_substitute = 0;	/* no SUBSTITUTE */
	    for(i=0;i<gfp->membercount;i++)
	    {
		if ( i != localindex )
		     (mfp+i)->status = FAULTY;
		for(j=NETWORK_1;j<NETWORKS;j++)
		{
		     if ( (mfp+i)->networks[j].use == USED )
	             {
		          (mfp+i)->networks[j].status = UNKNOWN;
		          (mfp+i)->networks[j].mstatus = UNKNOWN;
                     }
		     else 
		     {
		          (mfp+i)->networks[j].status = NA;
		          (mfp+i)->networks[j].mstatus = NA;
		     }
		     (mfp+i)->networks[j].flag = NETSTATUSRESET;
		     (mfp+i)->networks[j].reached = 1;
		     lib_timer_reset(&((mfp+i)->networks[j].validity));
		     lib_timer_reset(&((mfp+i)->networks[j].confirm));
		}
	    }
	    break;
	case WARM_REINIT : /* warm reinit */
	    break;
	}
}

/* void wat_threshold()
 * For each interface if the threshold flag is ENABLE then analyze if threshold
 * value has been reached.
 *
 * Argument : None
 *
 * return   : None
 */
void wat_threshold()
{
struct member    *m;
int               i;
int		  j;
int 		  k;

  for(i=0;i<MEMBER_COUNT;i++)
  {
      m = (mfp+i);
      for(j=0;j<NETWORKS;j++)
          if ( m->networks[j].use == USED )
      {
	  if ( m->networks[j].threshold == ENABLE )
	  {
	       k = m->validity - m->networks[j].value;
	       if ( lib_timer_reached(&(m->networks[j].validity),k) )
	       {
		   if ( !m->networks[j].reached )
		   {
		        sprintf(messagestring,
		   "Threshold %d reached for member %s on network %d (%s)",
		                m->networks[j].value,
	 	                m->name,
		                j+NETWORK_OFFSET,
	    	                m->networks[j].ifname);
		        PINFOR_B
		        messagestring
	  	        PINFOR_E;
                   }  
		   m->networks[j].reached++;
	       }
	  }
      }
      
  }
}

/* void wat_analyze_networks()
 * Analyze all interfaces to determine the status of each interface.
 *
 * 0 / Set number of alive systems
 * 1 / Analyze local member interface : 
 *	if validity flag reset && confirm flag reset -> status unchanged
 *	if validity flag set && validity event reset -> status ON
 *	if validity flag set && validity event set   
 *		alive >  BOTH -> status OFF
 *		alive <= BOTH -> status UNKNOWN
 *      if status != ON            -> network unavailable
 *      if status == ON            -> network available
 * 2 / Analyze remote member interface :
 *	if network unavailable                       -> status UNKNOWN
 *	if network available
 *	        if validity flag reset &&
 *                 confirm flag reset                -> status unchanged
 *	        if validity flag set && 
 *                 validity event reset              -> status ON
 *	        if validity flag set && 
 *                 validity event set   
 *                    inter network confirmation     -> status UNKNOWN || ON
 *	        if confirm  flag set && 
 *                 confirm event set                 
 *                    alive >  BOTH                  -> status OFF
 *                    alive >= BOTH                  -> status UNKNOWN
 *
 * Argument : None
 *
 * Return   : None
 */
void wat_analyze_networks()
{
struct member   *m;
int i;  /* index */
int j;	/* index */
int k;  /* index */
int alive; /* number of alive members */
int avail[NETWORKS]; /* interface flag */
struct wdx_timer *v;
struct wdx_timer *c;
struct eventscript_data evt; /* data used to call lib_eventscript */

#define UNAVAIL 0
#define AVAIL 1

/* init */
   for(j=0;j<NETWORKS;j++)
    avail[j] = UNAVAIL;
   for(k=0;k<gfp->membercount;k++)
       for(j=0;j<NETWORKS;j++)
	   (mfp+k)->networks[j].flag = NETSTATUSRESET;
/* 0 / Set number of alive systems */
   alive = how_many_members();
//printf("alive=%d\n", alive);

/* 1 / Analyze local member interface : */
   m = mfp + localindex;
   for(j=0;j<NETWORKS;j++)
        if(m->networks[j].use == USED && m->networks[j].flag == NETSTATUSRESET)
   {
        v = &(m->networks[j].validity);
        c = &(m->networks[j].confirm);
        /* if validity flag reset && confirm flag reset -> no status change */
	if(!lib_timer_alive(v) && !lib_timer_alive(c))
	{
//printf("!lib_timer_alive && !lib_timer_alive\n");
	    if ( alive > BOTH )
	    {
                 m->networks[j].flag = NETSTATUSSET;
    	         m->networks[j].mstatus = m->networks[j].status;
	         lib_timer_reset(c);
	         lib_timer_reset(v);
	    }
	    else
	    {
                 m->networks[j].flag = NETSTATUSSET;
	         m->networks[j].mstatus = m->networks[j].status;
		 m->networks[j].status = UNKNOWN;
//printf("COINNNCOIN\n");
	         lib_timer_reset(c);
	         lib_timer_reset(v);
	    }
	}

        /* if validity flag set && validity event reset -> status ON */
	if(lib_timer_valid(v))
	{
//printf("lib_timer_valid\n");
            m->networks[j].flag = NETSTATUSSET;
	    m->networks[j].mstatus = m->networks[j].status;
	    m->networks[j].status = ON;
	    lib_timer_reset(c);
	}
        /* if validity flag set && validity event set */
	if(lib_timer_event(v))
        {
//printf("lib_timer_event\n");
            /* alive >  BOTH -> status OFF */
	    if ( alive > BOTH )
	    {
                 m->networks[j].flag = NETSTATUSSET;
	         m->networks[j].mstatus = m->networks[j].status;
                 m->networks[j].status = OFF;
	         lib_timer_reset(c);
	         lib_timer_reset(v);
	    }
            /* alive <= BOTH -> status UNKNOWN */
	    else 
	    {
                 m->networks[j].flag = NETSTATUSSET;
	         m->networks[j].mstatus = m->networks[j].status;
                 m->networks[j].status = UNKNOWN;
//printf("COINNNCOIN2\n");
	         lib_timer_reset(c);
	         lib_timer_reset(v);
	    }
        }
        /* if status != ON -> network unavailable */
	if ( m->networks[j].status != ON )
	     avail[j] = UNAVAIL;
        /* if status == ON -> network available */
        else avail[j] = AVAIL;

   }
/* 2 / Analyze remote member interface : */
   for(k=0;k<gfp->membercount;k++)
       if ( k != localindex )
   {
       m = mfp + k;
       for(j=0;j<NETWORKS;j++)
           if(m->networks[j].use == USED && 
              m->networks[j].flag == NETSTATUSRESET)
       {
//printf("Analyze remote member interface\n");
	   v = &(m->networks[j].validity);
	   c = &(m->networks[j].confirm);
           switch(avail[j])
	   {
	   case UNAVAIL : /* if network unavailable -> status UNKNOWN */

		 if(member_confirmation(m,j))
		 {
		     for(i=0;i<NETWORKS;i++)
			if( m->networks[i].use == USED )
		     {
                         m->networks[i].flag = NETSTATUSSET;
			 m->networks[i].mstatus = m->networks[i].status;
			 m->networks[i].status = UNKNOWN;
			 lib_timer_reset(v);
			 lib_timer_reset(c);
		     }
		 }
		 else
		 {
                     m->networks[j].flag = NETSTATUSSET;
    	             m->networks[j].mstatus = m->networks[j].status;
                     m->networks[j].status = UNKNOWN;
	             lib_timer_reset(c);
	             lib_timer_reset(v);
                 }
		 break;	    
	   case AVAIL : /* if network available */
                 /* if validity flag reset && */
                 /*    confirm flag reset                -> no change */
	         if(!lib_timer_alive(v) && !lib_timer_alive(c))
	         {
                        m->networks[j].flag = NETSTATUSSET;
  	                m->networks[j].mstatus = m->networks[j].status;
	                lib_timer_reset(c);
	                lib_timer_reset(v);
	         }

                 /* if validity flag set &&  */
                 /*    validity event reset              -> status ON */
		 if(lib_timer_valid(v))
		 {
                       m->networks[j].flag = NETSTATUSSET;
                       m->networks[j].mstatus = m->networks[j].status;
                       m->networks[j].status = ON;
	               lib_timer_reset(c);
		 }
                 /* if validity flag set &&  */
                 /*    validity event set    */
                 /*    inter network confirmation     -> status UNKNOWN || ON */
		 if(lib_timer_event(v))
		 {
			if(member_confirmation(m,j))
			{
			    for(i=0;i<NETWORKS;i++)
				if(m->networks[i].use == USED )
			    {
                                m->networks[i].flag = NETSTATUSSET;
				m->networks[i].mstatus = m->networks[i].status;
				m->networks[i].status = UNKNOWN;
				lib_timer_reset(v);
				lib_timer_reset(c);
			    }
			}
			else
			{
                            m->networks[j].flag = NETSTATUSSET;
	                    m->networks[j].mstatus = m->networks[j].status;
                            m->networks[j].status = ON;
	                    lib_timer_reset(v);
	                    lib_timer_set(c,m->broadcast);
			
			}
		 }
                 /* if confirm  flag set && */
                 /*    confirm event set    */
	         if(lib_timer_event(c))
                 {
                      /* alive >  BOTH -> status OFF */
	               if ( alive > BOTH )
	               {
                            m->networks[j].flag = NETSTATUSSET;
	                    m->networks[j].mstatus = m->networks[j].status;
                            m->networks[j].status = OFF;
	                    lib_timer_reset(c);
	                    lib_timer_reset(v);
	               }
                       /* alive <= BOTH -> status UNKNOWN */
	               else 
	               {
                            m->networks[j].flag = NETSTATUSSET;
	                    m->networks[j].mstatus = m->networks[j].status;
                            m->networks[j].status = UNKNOWN;
	                    lib_timer_reset(c);
	                    lib_timer_reset(v);
	               }
                 }
		 break;
	   } 
       }
   }

   /* network transitions */
   for(k=0;k<gfp->membercount;k++)
   {
	m = mfp + k;
        for(j=0;j<NETWORKS;j++)
        {
     	     if ( m->networks[j].use == USED )
	          if ( m->networks[j].mstatus != m->networks[j].status )
             {		       
		  evt.type = EVT_NETWORK; 
		  evt.display = YES;
		  evt.script = watchdogactivity;
		  evt.index = k;
		  evt.group = gfp;
		  evt.data.n.net = j;
		  evt.data.n.old = m->networks[j].mstatus;
		  evt.data.n.new = m->networks[j].status;
		  lib_eventscript(&evt);     
	     }
        }
   }
}

/* int wat_input()
 * A watchdog input has been received. Update the status of remote and
 * local networks
 *
 * Argument : pt to watchdog structure
 *            network number on which the message was received

 *
 * Return   : 0 ok -1 fatal error
 */
int wat_input(w,a)
struct watchdog *w;
int a;
{

   if(w && a!= -1)
   {
       /* update sender network database only if local is in use*/
       if((mfp+localindex)->networks[a].use != USED )
	   return(0);
       /* update sender network database */
       input_net(w->index,w,a);

       /* update local interface status */
       input_net(localindex,w,a);

   }
   else
   {
       sprintf(messagestring,
               "Message received from not configured network %d",
               a);
       PERROR_B
       messagestring
       PERROR_E;
   }
}

/* void wat_activity_set
 * Set watchdogactivity(silence) and watchdogautomatic(lock) to the values 
 * passed as argument
 *
 */
void wat_activity_set(lockval,silval)
enum mask lockval;
enum mask silval;
{
   /* set static activity value */
   watchdogautomatic = lockval;
   watchdogactivity = silval;
}

/* void wat_validate_networks()
 * Validate the networks by decrementing by decrement all the wdx_timer set
 *
 * Argument : decrementation value
 */
void wat_validate_networks(decrement)
int decrement;    /* decrementation value */
{
int k;
int j;

   /* automatic get mastertoken priority delay */
   lib_timer_decrease(&autodelay,decrement);

/* decrements each wdx_timer by decrement */
   for(k=0;k<gfp->membercount;k++)
   {
        for(j=0;j<NETWORKS;j++)
        {
	    lib_timer_decrease(&((mfp+k)->networks[j].validity),decrement);
	    lib_timer_decrease(&((mfp+k)->networks[j].confirm),decrement);
        }
   }
}

/* int wat_local_member_transition()
 * Determine the status of the local member and detects local transitions
 *
 * Argument : kernel status
 *
 * Return : 0 ok / -1 exception
 */
/*int wat_local_member_transition(stat)
enum watchdog_status stat;
*/
int wat_local_member_transition(stat,brdflgpt)  /* MK001 */
enum watchdog_status stat;
int *brdflgpt;                                  /* MK001 */

{
struct member *m;               /* pointer to local member structure */
int       masown;               /* index of the mastertoken owner */
enum resource_status status;
enum member_status s;
int ret;

    /* init */
    ret = 0;
    m = mfp + localindex;    
    s = m->status;              /* store current status */
    m->status = IDLE;           /* default value */

    /* according to stat values */
    switch(stat)
    {
    case W_IDLE : /* watchdog idle status */
	break;
    case W_RUNNING : /* watchdog running status */
        /* master / substitute */
        if (m->type == MASTER_SUBSTITUTE_T) 
        {
             masown = lib_requester_of_resource(gfp->resourcelist,mastertoken);
	     status = lib_status_of_resource(gfp->resourcelist,mastertoken);
             switch(status)
             {
             case RESERVED :  /* MASTERTOKEN reserved */
                  if ( masown == localindex )
                  {
                       if ( s == SUBSTITUTE )
		       {	    
                            trigger_script(gfp->newmasterscript,NULL,NULL);
                            lck_purge_resource_delayed();
		       }
		       m->status = MASTERHOLDER;
                  }
                  else 
                  {
                       if ( s == MASTERHOLDER )		 
                       {
                            m->status = SUBSTITUTE;
                            ret = wat_node_exception(DOUBLE_MASTER,
                                          "Double master detected");
                       }
                       else m->status = SUBSTITUTE;
                  }
                  break;
             default :       /* MASTERTOKEN not RESERVED */
                  m->status = SUBSTITUTE;
                  break;            
             } /* end switch */
        } /* end master/substitute capabilities */
        else m->status = SATELLITE; /* it is considered to be a satellite */
	break;
    }
    /* message */
    if ( s != m->status )
    {     member_transition(localindex,s,m->status); 
         *brdflgpt =  *brdflgpt + 1;         /* MK001 */
    }
    return(ret);
}

/* int wat_remote_member_transition()
 * For each remote member in the group, check the members' transitions.
 * Update the member status and trigger the different scripts, in case of
 * status change.
 *
 * Argument : None
 *
 * Return   : 0
 */
int wat_remote_member_transition()
{
int ind1;
int trans;
int own;
enum resource_status status;
struct member *m;

    for(ind1=0;ind1<gfp->membercount;ind1++)
    {
        if ( ind1 != localindex )
        {
             trans = 0;
             m = mfp + ind1;
             if ( lib_is_member_valid(mfp,ind1) )    /* member is valid */
             {
                  if ( m->status != m->brdstatus ) /* member transition */
                  {
                       if ( m->brdstatus == MASTERHOLDER )
                       {
                            own = lib_requester_of_resource(gfp->resourcelist,
                                    mastertoken);
                            status = lib_status_of_resource(gfp->resourcelist,
                                    mastertoken);
                            if ( status == RESERVED && own == ind1 )
			    {
				 lck_purge_resource_delayed();	 
                                 trans++;
                            }
                       }
                       else trans++;
                  }
                  if ( trans )
                  {
                       member_transition(ind1,m->status,m->brdstatus);
                       m->status = m->brdstatus;
                  }
             }
             else                            /* member is not valid */
             {
                  if ( m->status != FAULTY ) /* member transition */
                       {
                       member_transition(ind1,m->status,FAULTY);
                       lck_purge_requester(ind1,
					   m->status,
					   maxpriority*automatictimer );
                       }
             m->status = FAULTY;
             }
        }
    }
    return(0);
}

/* void wat_failure()
 * Performs the watchdog failure protocol
 *
 * Argument : Local status
 *	      Pt to broadcast flag
 *
 * Return   : None
 */
void wat_failure(stat,brdflgpt)
enum watchdog_status stat;
int *brdflgpt;
{
int                   tokenowner;
enum resource_status tokenstatus;
int         there_is_now_a_substitute;
int                          inx;

    tokenstatus = lib_status_of_resource(gfp->resourcelist,mastertoken);
    tokenowner = lib_requester_of_resource(gfp->resourcelist,mastertoken);
    switch ((mfp+localindex)->status)
    {
   
    /* Try to become master if the following requirements are fulfilled	*/
    /* (R1) The local node must be already a substitute  			*/
    /* (R2) The local node must not have requested the mastertoken yet	*/
    /* (R3) No other node has already a lock on the mastertoken		*/

    case SUBSTITUTE:							/* (R1) */
       if ( tokenstatus == FREE &&
/* masterowner != NOT_AN_OWNER && */
	    !lib_timer_alive(&autodelay) && !automaticflag &&
	    watchdogautomatic == YES )
       {
	    automaticflag++;
	    if ( masterowner != localindex )
		 lib_timer_set(&autodelay,localpriority*automatictimer);
	    else lib_timer_set(&autodelay,maxpriority*automatictimer);
       }
       if ( tokenstatus == FREE && lib_timer_event(&autodelay) )
       {
	    /* (R2) and (R3) */
            if ( lck_automatic_request_wish(mastertoken,brdflgpt) != 0 )
            {		 
		 /* wait for next retry */
		 automaticflag++;
		 lib_timer_set(&autodelay,localpriority*automatictimer); 
	    }
            else lib_timer_reset(&autodelay);
       }
       masterowner = tokenowner; /* remember current master owner */
       break;

    /* Satellites must watchdog the master for sudden disappearance	*/
    /* Detection of master disappearance is as follows			*/
    /* (S1) The local node must be already a satellite and 		*/
    /* (S2) The index of the master node changed from defined		*/
    /* (S3) to undefined	or					*/
    /* (S4) to another defined index.					*/

    case SATELLITE:                                             /* (S1) */
       if (masterowner != NOT_AN_INDEX)                         /* (S2) */
       {
           if ( tokenstatus == FREE )                           /* (S3) */
	   {
                trigger_script(gfp->satellitescript,MASTER_DISAPPEARENCE,
                                   "Satellite detects master disappearence");
	        masterowner = NOT_AN_INDEX;
	   }
           else if ( masterowner != tokenowner )                /* (S4) */
	   {
                trigger_script(gfp->satellitescript,MASTER_CHANGE,
                                   "Satellite detects master change");
		masterowner = tokenowner;
           }
       }
       else if ( tokenstatus == RESERVED )
	   masterowner = tokenowner;
       break;

    /* ***************** NO MORE SUBSTITUTE DETECTION ************************/
    /* The Masterholder must trigger NOSUBSTITUTESCRIPT if it detects        */
    /* that no more substitutes are available in the group.                  */
    /*      Trigger, if 						     */
    /* (M1) The local node is a masterholder, and		             */
    /* (M4) There was at least a substitute in the group                     */
    /* (M5) but now, there are no more substitutes                           */

    case MASTERHOLDER:						/* (M1)	*/
       /* Search for a substitute */
       for (there_is_now_a_substitute=0, inx=0; inx < gfp->membercount; inx++)
       {
            if ((lib_is_member_valid(mfp,inx)) && ((mfp+inx)->status == SUBSTITUTE))
            {
                 there_is_now_a_substitute = 1;
                 break;
            }
       }

       if ( (there_was_a_substitute) &&                     /* (M4) */
            ( ! there_is_now_a_substitute ) )               /* (M5) */
            trigger_script(gfp->nosubstitutescript,NULL,
                               "Masterholder detects all substitutes disappearence");

       /* if we have found a substitute, remember it */
       there_was_a_substitute = there_is_now_a_substitute ; 
       masterowner = tokenowner;
       break;
    
    default: /* IDLE */
       break;
    }
    if ( tokenstatus != FREE )
    {
	lib_timer_reset(&autodelay);
	automaticflag = 0;
    }
}

/* int wat_node_exception()
 * Procedure used to trigger exception script
 *
 * Argument : Pt to argument string
 *            Pt to message string
 *
 * Return : -1 to stop the execution of wdx
 */			
int wat_node_exception(a,s)
char *a;
char *s;
{

    trigger_script(gfp->exceptionscript,a,s); 
    return(0);
}

/*
 * LOCAL ROUTINES : no prefixe
 */

/* void trigger_script
 * Local routine used to filter script triggering in case of watchdogactivity set
 *
 * Argument : pt to script
 *	      pt to argument
 *	      pt to message
 *
 * Return : None
 */
void trigger_script(s,a,m)
char *s;    /* pt to script */
char *a;    /* pt to argument */
char *m;    /* pt to message */
{
char              st[STRING_LEN];
char              *sc;

    if ( *s != '\0' )
    {    
	if ( a )
	{
	     sprintf(st,"%s %s",s,a);
             sc = st;
	}
	else sc = s;
    }
    else sc = NULL;

    if(watchdogactivity == YES )
       spc_trigger_script(sc,m);
}

/* void member_transition()
 * Trigger the Eventscript for a member transition
 * The eventscript is triggered according to Functionnal Specs.
 * The severity of the message is INFORMATION or ERROR according to 
 * Functionnal Specs.
 *
 * Argument : member index
 *            old status
 *            new status
 *
 * Return   : none
 */
void member_transition(mi,os,ns)
int mi;
enum member_status os;
enum member_status ns;
{
struct eventscript_data evt; /* data used to call lib_eventscript */

	evt.type = EVT_MEMBER; 
	evt.display = YES;
	evt.script = watchdogactivity;
	evt.index = mi;
	evt.group = gfp;
	evt.data.m.old = os;
	evt.data.m.new = ns;
	lib_eventscript(&evt);     
}

/* int how_many_members()
 * Return a value > BOTH if more than two members available
 *
 * Argument : None
 *
 * Return : number of available members.
 */
int how_many_members()
{
struct member *_m;
int _i;
int _j;
int _ret;
int _a;

    _ret = 0;
    for(_i=0;_i<gfp->membercount;_i++)
    {
	_m = mfp+_i;
	_a = 0;
        for(_j=0;_j<NETWORKS;_j++)
        {
	    if(_m->networks[_j].use == USED )
	    {
               if(lib_timer_valid(&(_m->networks[_j].validity)))
		  _a++;
            }
        }
	if(_a)
	   _ret++;
    }
    return(_ret);
}

/* int is_multi_networks()
 * Return a value > SINGLE if more than one interface defined for this member
 *
 * Argument : pt to member structure
 */
int is_multi_networks(m)
struct member *m;
{
int _i;
int _ret;

    _ret = 0;
    for(_i=0;_i<NETWORKS;_i++)
        if(m->networks[_i].use == USED )
 	    _ret++;
    return(_ret);
}

/* int member_confirmation()
 * For a member and a specified network, check if other interface are in
 * confirmation state or validity expired to determine if member is really 
 * down or not.
 *
 * Argument : pt to member structure
 *            network number
 *
 * Return : >0 if member down / 0 otherwise
 */
int member_confirmation(m,i)
struct member *m;
int            i;
{
int j;
int k;
int l;
int f;

    f = 0;
    if ( is_multi_networks(m) > SINGLE )
    {
        k = 0;
        l = 0;
        for(j=0;j<NETWORKS;j++)
            if(j!=i)
        {
            /* count of used interfaces */
            if( m->networks[j].use == USED )
		k++;
            /* count of used interfaces in OFF, UNKNWON state or in 
               confirmation/event */
            if( m->networks[j].use == USED &&
/*	        ( m->networks[j].status == OFF ||
	          ( m->networks[j].status == UNKNOWN && 
                    m->networks[j].validity.flag == FLAGRESET ) ||
		  m->networks[j].validity.event == EVENTSET ||
	          m->networks[j].confirm.flag == FLAGSET ) )
*/
		
	        ( !lib_timer_alive(&m->networks[j].validity) ||
		   lib_timer_event(&m->networks[j].validity) ||
	           lib_timer_alive(&m->networks[j].confirm)  ) )
	        l++;
            if ( l == k ) /* all used interfaces in OFF, UNKNOWN, confirm */
                 f++;
            else f = 0;
        }
    }
    else f++;
    return(f);
}

/* int input_net()
 * Update a member network table in case of message input
 * If a message is received on one interface and multi networks are defined
 * for local and remote members, then confirmation timeout can be started to
 * allow transition from UNKNOWN to OFF for these interfaces.
 *
 * Argument : member index 
 *	      pt to watchdog structure
 *            network number
 */
int input_net(n,w,i)
int              n;
struct watchdog *w;
int              i;
{
struct member   *m;
int j;	/* index */

   m = mfp+n;
   switch(m->networks[i].status)
   {
   case NA :      /* update */
	lib_timer_reset(&(m->networks[i].confirm));
	break;
   case UNKNOWN : /* global + single update */
	lib_timer_reset(&(m->networks[i].confirm));
	break;
   case ON :  /* single update */
	/* no transition */
	lib_timer_reset(&(m->networks[i].confirm));
	break;
   case OFF : /* single update transition */
	lib_timer_reset(&(m->networks[i].confirm));
	break;
   }

   /* associated remote interfaces */
   if ( is_multi_networks(m) > SINGLE && 
        is_multi_networks(mfp+localindex) > SINGLE )
   {
        for(j=0;j<NETWORKS;j++)
            if(m->networks[j].use == USED &&
               m->networks[j].status == UNKNOWN && j != i && 
               m->networks[j].confirm.flag == FLAGRESET )
            {
               /* set confirmation delay for remote interface only */
               if ( n != localindex )
               {
       	            lib_timer_set(&(m->networks[j].confirm),m->broadcast);
                    lib_timer_reset(&(m->networks[j].validity));
                }
            }
   }

   /* update status */
   if ( w->validity > m->networks[i].validity.value)
        lib_timer_set(&(m->networks[i].validity),w->validity);
   m->networks[i].reached = 0;
   /* store received status value */
   m->brdstatus = w->status;  
  
   return(0);
}
